import weave
from weave.legacy.weave import context_state as _context
from weave.legacy.weave import graph
from weave.legacy.weave import weave_types as types
from weave.legacy.weave.weave_internal import make_const_node

from ...legacy.weave import registry_mem

_loading_builtins_token = _context.set_loading_built_ins()


@weave.op()
def _test_add_one(x: int) -> int:
    return x + 1


@weave.op(
    input_type={"x": types.TypedDict({"a": types.Any()})},
    output_type=lambda input_type: input_type["x"].property_types["a"],
)
def _test_mappability_custom_pick_op(x):
    return x["a"]


@weave.op(
    input_type={"obj": types.TypedDict({}), "key": types.String()},
    output_type=types.TypeType(),
)
def _test_mappability_custom_pick_refine(obj, key):
    return types.TypeRegistry.type_of(obj[key])


@weave.op(
    input_type={"obj": types.TypedDict({}), "key": types.String()},
    output_type=types.Any(),
    refine_output_type=_test_mappability_custom_pick_refine,
)
def _test_mappability_custom_pick_op_with_refine(obj, key):
    return obj[key]


_context.clear_loading_built_ins(_loading_builtins_token)


def test_basic_mapping():
    val = weave.save([1, 2, 3])
    assert weave.use(val + 1) == [2, 3, 4]
    assert weave.use(val - 1) == [0, 1, 2]

    val = weave.save(["1", "2", "3"])
    assert weave.use(val + "1") == ["11", "21", "31"]

    val = weave.save(["a", "b", "c"])
    assert weave.use(val.upper()) == ["A", "B", "C"]


def test_non_mapped_use():
    node = weave.save(1)._test_add_one()
    assert node.type == weave.types.Int()
    assert weave.use(node) == 2


def test_non_mapped_serialized():
    node = weave.legacy.weave.weave_internal.make_output_node(
        weave.types.Int(),
        _test_add_one.name,
        {"x": weave.legacy.weave.graph.ConstNode(weave.types.Int(), 1)},
    )
    assert weave.use(node) == 2


def test_mapped_use():
    node = weave.save([1, 2, 3])._test_add_one()
    assert node.type == weave.types.List(weave.types.Int())
    assert weave.use(node) == [2, 3, 4]


def test_mapped_nullable_use():
    node = weave.save([1, None, 3])._test_add_one()
    assert node.type == weave.types.List(weave.types.optional(weave.types.Int()))
    assert weave.use(node) == [2, None, 4]


def test_mapped_serialized():
    node = weave.legacy.weave.weave_internal.make_output_node(
        weave.types.Int(),
        _test_add_one.name,
        {
            "x": weave.legacy.weave.graph.ConstNode(
                weave.types.List(weave.types.Int()), [1, 2, 3]
            )
        },
    )
    assert weave.use(node) == [2, 3, 4]


def test_mapped_empty_use():
    node = weave.save([])._test_add_one()
    assert node.type == weave.types.List(weave.types.Int())
    assert weave.use(node) == []


def test_mapped_empty_serialized():
    node = weave.legacy.weave.weave_internal.make_output_node(
        weave.types.Int(),
        _test_add_one.name,
        {"x": weave.legacy.weave.graph.ConstNode(weave.types.List(weave.types.Int()), [])},
    )
    assert weave.use(node) == []


def test_custom_class():
    _loading_builtins_token = _context.set_loading_built_ins()

    @weave.type()
    class TestType:
        @weave.op()
        def test_fn(self, a: int) -> int:
            return a + 1

    _context.clear_loading_built_ins(_loading_builtins_token)

    node = TestType().test_fn(1)
    assert weave.use(node) == 2

    node_list = weave.legacy.weave.ops.make_list(**{"0": TestType(), "1": TestType()})
    node = node_list.test_fn(1)
    assert weave.use(node) == [2, 2]


def test_pick_index_challenge():
    node = weave.save([{"a": 1}])
    index_pick = node[0]["a"]
    assert weave.use(index_pick) == 1

    pick_index = node["a"][0]
    assert weave.use(pick_index) == 1


def test_mapped_maybe_list():
    a = weave.save([3, None, 4, None, 5])
    res = a + 1
    assert res.type == types.List(types.union(types.Number(), types.NoneType()))
    assert weave.use(res) == [4, None, 5, None, 6]


def test_mapped_maybe_custom_pick():
    a = weave.save([{"a": 3}, None, {"a": 4}, None, {"a": 5}])
    res = a._test_mappability_custom_pick_op()
    assert res.type == types.List(types.union(types.Int(), types.NoneType()))
    assert weave.use(res) == [3, None, 4, None, 5]


def test_mapped_maybe_pick():
    a = weave.save([{"a": 3}, None, {"a": 4}, None, {"a": 5}])
    res = a.pick("a")
    assert res.type == types.List(types.union(types.Int(), types.NoneType()))
    assert weave.use(res) == [3, None, 4, None, 5]


def test_mapped_maybe_custom_refine():
    a = weave.save([{"a": 3}, None, {"a": 4}, None, {"a": 5}])
    res = a._test_mappability_custom_pick_op_with_refine("a")
    assert res.type == types.List(types.union(types.Int(), types.NoneType()))
    assert weave.use(res) == [3, None, 4, None, 5]



import os

import pytest

import weave
import weave.legacy.weave.context_state
import weave.legacy.weave.wandb_api
import weave.trace.weave_init


def test_create_list_rename_delete():
    os.environ["WEAVE_CACHE_MODE"] = "minimal"

    # create
    art_node = weave.legacy.weave.ops.get("local-artifact:///my-data:latest/obj")
    art_node.set("foo")
    assert weave.use(art_node) == "foo"

    # list
    arts = weave.use(weave.legacy.weave.ops.local_artifacts())
    assert len(arts) == 1
    assert arts[0].name == "my-data"

    # rename
    weave.legacy.weave.ops.rename_artifact(art_node, "my-data2")
    arts = weave.use(weave.legacy.weave.ops.local_artifacts())
    assert len(arts) == 1
    assert arts[0].name == "my-data2"

    # delete
    art_node = weave.legacy.weave.ops.get("local-artifact:///my-data2:latest/obj")
    weave.legacy.weave.ops.delete_artifact(art_node)
    arts = weave.use(weave.legacy.weave.ops.local_artifacts())
    assert len(arts) == 0




import importlib
import typing

import weave
from weave.trace.op_extensions.accumulator import add_accumulator
from weave.trace.patcher import MultiPatcher, SymbolPatcher

if typing.TYPE_CHECKING:
    from mistralai.models.chat_completion import (
        ChatCompletionResponse,
        ChatCompletionStreamResponse,
    )


def mistral_accumulator(
    acc: typing.Optional["ChatCompletionResponse"],
    value: "ChatCompletionStreamResponse",
) -> "ChatCompletionResponse":
    # This import should be safe at this point
    from mistralai.models.chat_completion import (
        ChatCompletionResponse,
        ChatCompletionResponseChoice,
        ChatMessage,
    )
    from mistralai.models.common import UsageInfo

    if acc is None:
        acc = ChatCompletionResponse(
            id=value.id,
            object=value.object,
            created=value.created,
            model=value.model,
            choices=[],
            usage=UsageInfo(prompt_tokens=0, total_tokens=0, completion_tokens=None),
        )

    # Merge in the usage info
    if value.usage is not None:
        acc.usage.prompt_tokens += value.usage.prompt_tokens
        acc.usage.total_tokens += value.usage.total_tokens
        if acc.usage.completion_tokens is None:
            acc.usage.completion_tokens = value.usage.completion_tokens
        else:
            acc.usage.completion_tokens += value.usage.completion_tokens

    # Loop through the choices and add their deltas
    for delta_choice in value.choices:
        while delta_choice.index >= len(acc.choices):
            acc.choices.append(
                ChatCompletionResponseChoice(
                    index=len(acc.choices),
                    message=ChatMessage(role="", content=""),
                    finish_reason=None,
                )
            )
        acc.choices[delta_choice.index].message.role = (
            delta_choice.delta.role or acc.choices[delta_choice.index].message.role
        )
        acc.choices[delta_choice.index].message.content += (
            delta_choice.delta.content or ""
        )
        if delta_choice.delta.tool_calls:
            if acc.choices[delta_choice.index].message.tool_calls is None:
                acc.choices[delta_choice.index].message.tool_calls = []
            acc.choices[
                delta_choice.index
            ].message.tool_calls += delta_choice.delta.tool_calls
        acc.choices[delta_choice.index].finish_reason = (
            delta_choice.finish_reason or acc.choices[delta_choice.index].finish_reason
        )

    return acc


def mistral_stream_wrapper(fn: typing.Callable) -> typing.Callable:
    op = weave.op()(fn)
    acc_op = add_accumulator(op, lambda inputs: mistral_accumulator)  # type: ignore
    return acc_op


mistral_patcher = MultiPatcher(
    [
        # Patch the sync, non-streaming chat method
        SymbolPatcher(
            lambda: importlib.import_module("mistralai.client"),
            "MistralClient.chat",
            weave.op(),
        ),
        # Patch the sync, streaming chat method
        SymbolPatcher(
            lambda: importlib.import_module("mistralai.client"),
            "MistralClient.chat_stream",
            mistral_stream_wrapper,
        ),
        # Patch the async, non-streaming chat method
        SymbolPatcher(
            lambda: importlib.import_module("mistralai.async_client"),
            "MistralAsyncClient.chat",
            weave.op(),
        ),
        # Patch the async, streaming chat method
        SymbolPatcher(
            lambda: importlib.import_module("mistralai.async_client"),
            "MistralAsyncClient.chat_stream",
            mistral_stream_wrapper,
        ),
    ]
)



import wandb

import weave


def test_basic_table(fake_wandb):
    table = wandb.Table(data=[[1, 2, 3]], columns=["a", "b", "c"])

    art = wandb.Artifact("test_name", "test_type")
    art.add(table, "table")
    art_node = fake_wandb.mock_artifact_as_node(art)

    file_node = art_node.file("table.table.json")
    table_node = file_node.table()
    table_rows = table_node.rows()

    raw_data = weave.use(table_rows).to_pylist_notags()
    assert raw_data == [
        {"a": 1, "b": 2, "c": 3},
    ]



import dataclasses
import typing

import weave
from weave.legacy.weave import weave_internal
from weave.legacy.weave import panel
from weave.legacy.weave.panels import panel_group


@dataclasses.dataclass
class BoardPanelLayout:
    x: int
    y: int
    h: int
    w: int


@dataclasses.dataclass
class BoardPanel:
    panel: typing.Any
    id: typing.Optional[str] = None
    layout: typing.Optional[BoardPanelLayout] = None


def varbar(editable=True, items=None) -> panel_group.Group:
    showExpressions = True if editable else "titleBar"
    if items is None:
        items = {}
    return panel_group.Group(
        config=panel_group.GroupConfig(
            layoutMode="vertical",
            equalSize=False,
            style="width:300px;",
            showExpressions=showExpressions,  # type: ignore
            allowedPanels=[
                "Expression",
                "Query",
                "Slider",
                "StringEditor",
                "SelectEditor",
                "Dropdown",
                "DateRange",
                "FilterEditor",
                "GroupingEditor",
            ],
            enableAddPanel=editable,
            childNameBase="var",
        ),
        items=items,
    )


def main() -> panel_group.Group:
    return weave.legacy.weave.panels.Group(
        layoutMode="grid",
        showExpressions=True,
        enableAddPanel=True,
    )


def Board(
    vars: typing.Union[panel_group.Group, list, dict],
    panels: typing.Union[panel_group.Group, list[BoardPanel]],
    editable=True,
):
    showExpressions = True if editable else "titleBar"
    vb = vars
    if not isinstance(vb, weave.legacy.weave.panels.Group):
        vb = varbar(editable=editable, items=vars)

    main = panels
    if not isinstance(panels, weave.legacy.weave.panels.Group):
        main_items = {}
        main_panel_layouts: list[panel_group.LayedOutPanel] = []
        for i, p in enumerate(panels):
            id = p.id
            if id is None:
                id = "panel" + str(i)
            main_items[id] = p.panel
            if p.layout is not None:
                main_panel_layouts.append(
                    panel_group.LayedOutPanel(
                        id=id,
                        layout=panel_group.LayoutParameters(
                            x=p.layout.x,
                            y=p.layout.y,
                            w=p.layout.w,
                            h=p.layout.h,
                        ),
                    )
                )
            main = panel_group.Group(
                config=panel_group.GroupConfig(
                    layoutMode="grid",
                    showExpressions=showExpressions,  # type: ignore
                    enableAddPanel=True,
                    disableDeletePanel=True,
                    gridConfig=panel_group.PanelBankSectionConfig(
                        id="grid0",
                        name="Section 0",
                        panels=main_panel_layouts,
                        isOpen=True,
                        flowConfig=panel_group.PanelBankFlowSectionConfig(
                            snapToColumns=True,
                            columnsPerPage=3,
                            rowsPerPage=2,
                            gutterWidth=10,
                            boxWidth=256,
                            boxHeight=256,
                        ),
                        type="grid",
                        sorted=0,
                    ),
                ),
                items=main_items,
            )

    result = panel_group.Group(
        config=panel_group.GroupConfig(
            layoutMode="horizontal",
            showExpressions=False,
        ),
        items={"sidebar": vb, "main": main},
    )
    result.finalize()
    return result



# Concat implementation for ArrowWeaveList. Should work in all cases!
#
# Any places where we extend lists need to follow our merge_types implementation.
# Fortunately, concat is the only place that extends lists! Everything else should
# make use of concat to do so.


import dataclasses
import typing

import numpy as np
import pyarrow as pa
from pyarrow import compute as pc

from weave.legacy.weave import errors
from weave.legacy.weave import weave_types as types
from weave.legacy.weave.arrow.list_ import (
    ArrowWeaveList,
    ArrowWeaveListGeneric,
    is_list_arrowweavelist,
    is_object_arrowweavelist,
    is_ref_arrowweavelist,
    is_taggedvalue_arrowweavelist,
    is_typedict_arrowweavelist,
    make_vec_none,
    offsets_starting_at_zero,
    unsafe_awl_construction,
)
from weave.legacy.weave.language_features.tagging import tagged_value_type

DEBUG = False

from weave.legacy.weave import engine_trace

tracer = engine_trace.tracer()


@dataclasses.dataclass
class UnionMember:
    values: ArrowWeaveList
    mask: pa.Array
    offsets: pa.Array


def indent_print(indent: int = 0, *args) -> None:
    if DEBUG:
        print("    " * indent + " ".join([str(a) for a in args]))


def pa_concat_arrays(arrays: list[typing.Union[pa.Array, pa.ChunkedArray]]) -> pa.Array:
    arrays = [a if isinstance(a, pa.Array) else a.combine_chunks() for a in arrays]

    with tracer.trace("pa_concat_arrays"):
        try:
            return pa.concat_arrays(arrays)
        except pa.lib.ArrowNotImplementedError as e:
            # pyarrow doesn't support concatenating all dictionaries, so we have to do it ourselves.
            if all(isinstance(a, pa.DictionaryArray) for a in arrays):
                # We can concatenate dictionaries by concatenating the indices and the dictionary
                # values separately.

                index_offset = 0
                indices = []
                for a in arrays:
                    indices.append(pc.add(a.indices, index_offset).cast(pa.int32()))
                    index_offset += len(a.dictionary)

                indices = pa_concat_arrays(indices)
                dictionary = pa_concat_arrays([a.dictionary for a in arrays])
                return pa.DictionaryArray.from_arrays(indices, dictionary)
            else:
                raise


def _concatenate_typeddicts(
    l1: ArrowWeaveListGeneric[types.TypedDict],
    l2: ArrowWeaveListGeneric[types.TypedDict],
    depth: int = 0,
) -> ArrowWeaveList:
    # Columns are in the order they're encountered.
    all_keys = {**l1.object_type.property_types, **l2.object_type.property_types}
    if not all_keys:
        # Both are empty
        return ArrowWeaveList(
            pa_concat_arrays([l1._arrow_data, l2._arrow_data]),
            types.TypedDict({}),
            invalid_reason="Possibly nullable",
        )
    properties: dict[str, ArrowWeaveList] = {}
    for key in all_keys:
        indent_print(depth, "TypedDict key", key)
        if key not in l1.object_type.property_types:
            properties[key] = _concatenate(
                make_vec_none(len(l1._arrow_data)), l2.column(key), depth=depth + 1
            )
        elif key not in l2.object_type.property_types:
            properties[key] = _concatenate(
                l1.column(key), make_vec_none(len(l2)), depth=depth + 1
            )
        else:
            properties[key] = _concatenate(
                l1.column(key), l2.column(key), depth=depth + 1
            )

    return ArrowWeaveList(
        pa.StructArray.from_arrays(
            [properties[key]._arrow_data for key in all_keys],
            list(all_keys),
            mask=pa_concat_arrays(
                [
                    pa.compute.is_null(l1._arrow_data),
                    pa.compute.is_null(l2._arrow_data),
                ]
            ),
        ),
        types.TypedDict({key: properties[key].object_type for key in all_keys}),
        l1._artifact,
        invalid_reason="Possibly nullable",
    )


def _concatenate_objects(
    l1: ArrowWeaveListGeneric[types.ObjectType],
    l2: ArrowWeaveListGeneric[types.ObjectType],
    depth: int = 0,
) -> ArrowWeaveList:
    assert l1.object_type.name == l2.object_type.name
    attrs: dict[str, ArrowWeaveList] = {}
    merged_type = typing.cast(
        types.ObjectType, types.merge_types(l1.object_type, l2.object_type)
    )
    for key in merged_type.property_types():
        attrs[key] = _concatenate(l1.column(key), l2.column(key), depth=depth + 1)
    return ArrowWeaveList(
        pa.StructArray.from_arrays(
            [attrs[key]._arrow_data for key in attrs],
            list(attrs.keys()),
            mask=pa_concat_arrays(
                [
                    pa.compute.is_null(l1._arrow_data),
                    pa.compute.is_null(l2._arrow_data),
                ]
            ),
        ),
        merged_type,
        l1._artifact,
        invalid_reason="Possibly nullable",
    )


def _concatenate_taggedvalues(
    l1: ArrowWeaveListGeneric[tagged_value_type.TaggedValueType],
    l2: ArrowWeaveListGeneric[tagged_value_type.TaggedValueType],
    depth: int = 0,
) -> ArrowWeaveList:
    tag = _concatenate(l1.tagged_value_tag(), l2.tagged_value_tag(), depth=depth + 1)
    value = _concatenate(
        l1.tagged_value_value(), l2.tagged_value_value(), depth=depth + 1
    )
    return ArrowWeaveList(
        pa.StructArray.from_arrays(
            [tag._arrow_data, value._arrow_data],
            ["_tag", "_value"],
            mask=pa_concat_arrays(
                [
                    pa.compute.is_null(l1._arrow_data),
                    pa.compute.is_null(l2._arrow_data),
                ]
            ),
        ),
        tagged_value_type.TaggedValueType(tag.object_type, value.object_type),  # type: ignore
        l1._artifact,
        invalid_reason="Possibly nullable",
    )


def convert_null_array_to_list(
    array: typing.Union[pa.ListArray, pa.NullArray],
) -> pa.ListArray:
    if isinstance(array, pa.NullArray):
        return pc.cast(array, pa.list_(pa.null()))
    else:
        return array


def _concatenate_lists(
    l1: ArrowWeaveListGeneric[types.List],
    l2: ArrowWeaveListGeneric[types.List],
    depth: int = 0,
) -> ArrowWeaveList:
    l1_arrow_data = convert_null_array_to_list(l1._arrow_data)
    l2_arrow_data = convert_null_array_to_list(l2._arrow_data)

    self_values: ArrowWeaveList = ArrowWeaveList(
        l1_arrow_data.flatten(), l1.object_type.object_type, l2._artifact
    )
    other_values: ArrowWeaveList = ArrowWeaveList(
        l2_arrow_data.flatten(), l2.object_type.object_type, l2._artifact
    )
    concatted_values = _concatenate(self_values, other_values, depth=depth + 1)
    new_offsets = pa_concat_arrays(
        [
            offsets_starting_at_zero(l1_arrow_data)[:-1],
            pa.compute.add(
                offsets_starting_at_zero(l2_arrow_data),
                len(self_values),
            ).cast(pa.int32()),
        ]
    )
    return ArrowWeaveList(
        pa.ListArray.from_arrays(
            new_offsets,
            concatted_values._arrow_data,
            mask=pa_concat_arrays(
                [
                    pa.compute.is_null(l1_arrow_data),
                    pa.compute.is_null(l2_arrow_data),
                ]
            ),
        ),
        types.List(
            types.merge_types(l1.object_type.object_type, l2.object_type.object_type)
        ),
        l1._artifact,
        invalid_reason="Possibly nullable",
    )


def _concatenate_non_unions(
    self: ArrowWeaveList, other: ArrowWeaveList, depth: int = 0
) -> typing.Optional[ArrowWeaveList]:
    if is_typedict_arrowweavelist(self) and is_typedict_arrowweavelist(other):
        return _concatenate_typeddicts(self, other, depth=depth)
    elif is_object_arrowweavelist(self) and is_object_arrowweavelist(other):
        if self.object_type.name == other.object_type.name:
            return _concatenate_objects(self, other, depth=depth)
    elif is_taggedvalue_arrowweavelist(self) and is_taggedvalue_arrowweavelist(other):
        return _concatenate_taggedvalues(self, other, depth=depth)
    elif is_list_arrowweavelist(self) and is_list_arrowweavelist(other):
        # Lists with different object types
        return _concatenate_lists(self, other, depth=depth)

    if isinstance(self.object_type, types.Number) and isinstance(
        other.object_type, types.Number
    ):
        # Numeric special case
        indent_print(
            depth, "Extend case number types", self.object_type, other.object_type
        )

        # Weave numeric types are odd for legacy reasons. We have three numeric types
        # Number, Float, Int. Unfortunately Number is a Union of Float and Int, but
        # also assignable to Int.

        # We should fix that. But for now we rely on the arrow type for this special
        # case.

        if self._arrow_data.type == other._arrow_data.type:
            new_arrow_type = self._arrow_data.type
        else:
            # If types aren't equal, just cast to float64. This is not entirely
            # correct! Very large integers will be truncated.
            new_arrow_type = pa.float64()

        new_weave_type: types.Type
        if pa.types.is_integer(new_arrow_type):
            new_weave_type = types.Int()
        elif pa.types.is_floating(new_arrow_type):
            new_weave_type = types.Float()
        else:
            raise errors.WeaveInternalError(
                f'Unexpected type "{new_arrow_type}" for "Number" type.'
            )

        if len(self) == 0:
            data = other._arrow_data
        elif len(other) == 0:
            data = self._arrow_data
        else:
            data = pa_concat_arrays(
                [
                    self._arrow_data.cast(new_arrow_type),
                    other._arrow_data.cast(new_arrow_type),
                ]
            )
        return ArrowWeaveList(
            data, new_weave_type, self._artifact, invalid_reason="Possibly nullable"
        )
    elif (
        # Types exactly equal case
        self.object_type == other.object_type
        or (
            # Types are refs, so arrow type is string
            is_ref_arrowweavelist(self)
            and is_ref_arrowweavelist(other)
            and self.object_type.__class__ == other.object_type.__class__
        )
    ):
        indent_print(depth, "Extend case equal", self.object_type, other.object_type)
        if len(self) == 0:
            data = other._arrow_data
        elif len(other) == 0:
            data = self._arrow_data
        else:
            data = pa_concat_arrays([self._arrow_data, other._arrow_data])
        return ArrowWeaveList(
            data,
            types.merge_types(self.object_type, other.object_type),
            self._artifact,
            invalid_reason="Possibly nullable",
        )
    return None


def _concatenate(
    self: "ArrowWeaveList", other: "ArrowWeaveList", depth=0
) -> "ArrowWeaveList":
    if depth > 50:
        raise ValueError("Maximum recursion depth exceeded")
    indent_print(depth, "EXTEND TOP", depth, self.object_type, other.object_type)

    # We use UnknownType for two things:
    #   1. to indicate that we don't know the type of an object,
    #   2. as the object type of a zero length array.
    # The former case, where we don't know the type of an object,
    # should never happen here! This is runtime and we know the types of
    # things.

    # Zero length arrays of unknown type
    if self.object_type == types.UnknownType():
        if len(self) == 0:
            return other
        else:
            raise errors.WeaveInternalError(
                'Encountered non-zero length "UnknownType" array'
            )
    if other.object_type == types.UnknownType():
        if len(other) == 0:
            return self
        else:
            raise errors.WeaveInternalError(
                'Encountered non-zero length "UnknownType" array'
            )

    # Cases where one of the types is None
    if isinstance(self.object_type, types.NoneType) and isinstance(
        other.object_type, types.NoneType
    ):
        return make_vec_none(len(self) + len(other))
    elif isinstance(self.object_type, types.NoneType):
        indent_print(depth, "Extend case self None")
        if len(self) == 0:
            return other
        elif len(other) == 0:
            return self
        self_data = pa.nulls(len(self), other._arrow_data.type)
        other_data = other._arrow_data
        data = pa_concat_arrays([self_data, other_data])
        return ArrowWeaveList(data, types.optional(other.object_type), other._artifact)
    elif isinstance(other.object_type, types.NoneType):
        indent_print(depth, "Extend case other None")
        if len(self) == 0:
            return other
        elif len(other) == 0:
            return self
        self_data = self._arrow_data
        other_data = pa.nulls(len(other), self._arrow_data.type)
        data = pa_concat_arrays([self_data, other_data])
        return ArrowWeaveList(data, types.optional(self.object_type), self._artifact)

    # Separate nulls from type
    self_nullable, self_non_none_type = types.split_none(self.object_type)
    other_nullable, other_non_none_type = types.split_none(other.object_type)
    self = self._with_object_type(
        self_non_none_type, invalid_reason="Possibly nullable"
    )
    other = other._with_object_type(
        other_non_none_type, invalid_reason="Possibly nullable"
    )
    result_nullable = self_nullable or other_nullable

    if not isinstance(self.object_type, types.UnionType) and not isinstance(
        other.object_type, types.UnionType
    ):
        result = _concatenate_non_unions(self, other, depth=depth)
        if result is not None:
            if result_nullable:
                return result._with_object_type(types.optional(result.object_type))
            else:
                return result._clear_invalid_reason()

    # Otherwise, we have types that can't be merged, we're going to produce a union

    if isinstance(self.object_type, types.UnionType):
        self_field_types = self.object_type.members
        self_type_codes = self._arrow_data.type_codes
        self_offsets = self._arrow_data.offsets
        self_fields = [
            self._arrow_data.field(i) for i in range(len(self._arrow_data.type))
        ]
    else:
        self_field_types = [self.object_type]
        self_type_codes = pa.array(np.zeros(len(self), dtype=np.int8))
        self_offsets = pa.array(np.arange(len(self), dtype=np.int32))
        self_fields = [self._arrow_data]

    if isinstance(other.object_type, types.UnionType):
        other_field_types = other.object_type.members
        other_type_codes = other._arrow_data.type_codes
        other_offsets = other._arrow_data.offsets
        other_fields = [
            other._arrow_data.field(i) for i in range(len(other._arrow_data.type))
        ]
    else:
        other_field_types = [other.object_type]
        other_type_codes = pa.array(np.zeros(len(other), dtype=np.int8))
        other_offsets = pa.array(np.arange(len(other), dtype=np.int32))
        other_fields = [other._arrow_data]

    assert len(self_field_types) == len(self_fields)
    assert len(other_field_types) == len(other_fields)

    # produce merged union
    new_member_awls: list[UnionMember] = []

    indent_print(depth, "MERGING UNIONS")
    indent_print(depth, "SELF NULLABLE", self_nullable)
    # indent_print(depth, "SELF MEMBERS")
    # for member in self_fields:
    #     indent_print(depth + 1, member)
    indent_print(depth, "OTHER NULLABLE", other_nullable)
    # indent_print(depth, "OTHER MEMBERS")
    # for member in other_fields:
    #     indent_print(depth + 1, member)

    remaining_other_indexes = {i: True for i in range(len(other_fields))}
    for self_i, (self_member_type, self_field) in enumerate(
        zip(self_field_types, self_fields)
    ):
        for other_i in remaining_other_indexes:
            other_member_type = other_field_types[other_i]
            other_field = other_fields[other_i]
            if types.types_are_mergeable(self_member_type, other_member_type):
                indent_print(
                    depth,
                    "SELF OTHER FIELD merge",
                    self_i,
                    other_i,
                    "self type:",
                    self_member_type,
                    "other type:",
                    other_member_type,
                )
                concatted = _concatenate_non_unions(
                    ArrowWeaveList(
                        self_field, self_member_type, invalid_reason="Possibly nullable"
                    ),
                    ArrowWeaveList(
                        other_field,
                        other_member_type,
                        invalid_reason="Possibly nullable",
                    ),
                )
                if concatted is None:
                    raise errors.WeaveInternalError(
                        "Could not concatenate mergable types"
                    )
                new_member_awls.append(
                    UnionMember(
                        values=concatted,
                        mask=pa_concat_arrays(
                            [
                                pa.compute.equal(self_type_codes, self_i).cast(
                                    pa.int8()
                                ),
                                pa.compute.equal(other_type_codes, other_i).cast(
                                    pa.int8()
                                ),
                            ]
                        ),
                        offsets=pa_concat_arrays(
                            [
                                self_offsets,
                                pa.compute.add(other_offsets, len(self_field)).cast(
                                    pa.int32()
                                ),
                            ]
                        ),
                    )
                )
                indent_print(
                    depth,
                    "SELF OTHER FIELD merge RESULT",
                    new_member_awls[-1].values.object_type,
                )
                remaining_other_indexes.pop(other_i)
                break
        else:
            # no matching other member found
            indent_print(depth, "SELF FIELD solo", self_i, self_member_type)
            new_member_awls.append(
                UnionMember(
                    values=ArrowWeaveList(
                        self_fields[self_i],
                        self_member_type,
                        self._artifact,
                        invalid_reason="Possibly nullable",
                    ),
                    mask=pa_concat_arrays(
                        [
                            pa.compute.equal(self_type_codes, self_i).cast(pa.int8()),
                            pa.array(np.zeros(len(other), dtype=np.int8)),
                        ]
                    ),
                    # offsets=pa.array(np.arange(len(self), dtype=np.int32)),
                    offsets=pa_concat_arrays(
                        [
                            self_offsets,
                            pa.array(np.zeros(len(other), dtype=np.int32)),
                        ]
                    ),
                )
            )
    for other_i in remaining_other_indexes:
        other_member_type = other_field_types[other_i]
        indent_print(depth, "OTHER FIELD solo", other_i, other_member_type)
        other_field = other_fields[other_i]
        new_member_awls.append(
            UnionMember(
                values=ArrowWeaveList(
                    other_field,
                    other_member_type,
                    other._artifact,
                    invalid_reason="Possibly nullable",
                ),
                mask=pa_concat_arrays(
                    [
                        pa.array(np.zeros(len(self), dtype=np.int8)),
                        pa.compute.equal(other_type_codes, other_i).cast(pa.int8()),
                    ]
                ),
                # offsets=pa.array(np.arange(len(other), dtype=np.int32)),
                offsets=pa_concat_arrays(
                    [
                        pa.array(np.zeros(len(self), dtype=np.int32)),
                        other_offsets,
                    ]
                ),
            )
        )

    # I think we will always have at least two union members here. Because we
    # must have had a union on one side coming in.
    if len(new_member_awls) < 2:
        raise errors.WeaveInternalError("Not enough union members after merge")

    member_types = [m.values.object_type for m in new_member_awls]

    result_type_codes = pa.array(np.zeros(len(self) + len(other), dtype=np.int8))
    for i, new_member_awl in enumerate(new_member_awls):
        result_type_codes = pa.compute.add(
            result_type_codes, pa.compute.multiply(new_member_awl.mask, i)
        )
    result_offsets = pa.array(np.zeros(len(self) + len(other), dtype=np.int32))
    for i, new_member_awl in enumerate(new_member_awls):
        result_offsets = pa.compute.add(
            result_offsets,
            pa.compute.multiply(new_member_awl.mask, new_member_awl.offsets),
        )

    assert len(member_types) == len(new_member_awls)
    expected_object_type_members = len(new_member_awls)
    new_object_type = types.UnionType(*member_types)
    if len(new_object_type.members) != expected_object_type_members:
        indent_print(
            depth,
            "ERROR PRODUCING UNION",
            len(new_object_type.members),
            len(new_member_awls),
        )
        indent_print(depth, "new_object.type.members")
        for m in new_object_type.members:
            indent_print(depth + 1, m)
        indent_print(depth, "new_member_awls")
        for m2 in new_member_awls:
            indent_print(depth + 1, m2.values.object_type)

        assert len(new_object_type.members) == expected_object_type_members

    indent_print(depth, "NEW MEMBER TYPE", new_object_type)
    indent_print(depth, "RESULT NULLABLE", result_nullable)
    return ArrowWeaveList(
        pa.UnionArray.from_dense(
            result_type_codes.cast(pa.int8()),
            result_offsets.cast(pa.int32()),
            [m.values._arrow_data for m in new_member_awls],
        ),
        types.optional(new_object_type) if result_nullable else new_object_type,
        self._artifact,
    )


def concatenate(
    self: "ArrowWeaveList", other: "ArrowWeaveList", depth=0
) -> "ArrowWeaveList":
    with unsafe_awl_construction("Possibly nullable"):
        result = _concatenate(self, other, depth)
    result.validate()
    return result


def _merge_concat_split(
    arr: list[ArrowWeaveList],
) -> tuple[list[ArrowWeaveList], list[ArrowWeaveList]]:
    if len(arr) < 2:
        raise ValueError("arr must have length of at least 2")
    middle_index = len(arr) // 2
    return arr[:middle_index], arr[middle_index:]


def concatenate_all(arr: list[ArrowWeaveList]) -> ArrowWeaveList:
    if len(arr) == 0:
        raise ValueError("arr must not be empty")
    if len(arr) == 1:
        return arr[0]
    left, right = _merge_concat_split(arr)
    return concatenate_all(left).concat(concatenate_all(right))



import asyncio

import pytest

from weave.legacy.weave import environment, errors, filesystem


@pytest.fixture()
def test_datafs():
    legacy_filesystem_dir = environment.weave_filesystem_dir
    environment.weave_filesystem_dir = lambda: "./testdata"
    yield
    environment.weave_filesystem_dir = legacy_filesystem_dir


def test_filesystem_access(test_datafs):
    fs = filesystem.Filesystem()
    assert fs.exists("spring-lamb.jpg")
    with pytest.raises(errors.WeaveAccessDeniedError):
        fs.exists("../spring-lamb.jpg")
    with pytest.raises(errors.WeaveAccessDeniedError):
        fs.exists("///spring-lamb.jpg")
    with pytest.raises(errors.WeaveAccessDeniedError):
        fs.exists("test_dir/../../weave")


@pytest.mark.asyncio()
async def test_filesystem_updates_atime_on_read(enable_touch_on_read, test_datafs):
    filename = "spring-lamb.jpg"
    fs = filesystem.FilesystemAsync()
    old_atime = (await fs.stat(filename)).st_atime
    async with fs.open_read(filename):
        await asyncio.sleep(0.5)  # give the spawned task time to execute
        new_atime = (await fs.stat(filename)).st_atime
    assert new_atime > old_atime

    fs = filesystem.Filesystem()
    old_atime = fs.stat(filename).st_atime
    with fs.open_read(filename):
        new_atime = fs.stat(filename).st_atime
    assert new_atime > old_atime



import typing

from weave.legacy.weave import weave_types as types
from weave.legacy.weave import op_def
from weave.legacy.weave.arrow.arrow import ArrowWeaveListType
from weave.legacy.weave.language_features.tagging import tagged_value_type

from .decorator_op import op

TypeCallable = typing.Callable[..., types.Type]
TypeOrCallable = typing.Union[types.Type, TypeCallable]
InputTypeDict = typing.Mapping[str, TypeOrCallable]


def _nullify_vector_type(t: types.Type) -> types.Type:
    if isinstance(t, ArrowWeaveListType):
        return ArrowWeaveListType(types.optional(t.object_type))
    elif isinstance(t, types.List):
        return types.List(types.optional(t.object_type))
    elif isinstance(t, tagged_value_type.TaggedValueType):
        return tagged_value_type.TaggedValueType(t.tag, _nullify_vector_type(t.value))
    elif isinstance(t, types.UnionType):
        return types.UnionType(*[_nullify_vector_type(mem) for mem in t.members])
    else:
        return types.optional(t)


def _make_nullable_vector_input_type_callable(
    old_type_func: TypeCallable,
) -> TypeCallable:
    def new_callable_nullable_awl_type(
        non_callable_input_types: dict[str, types.Type]
    ) -> types.Type:
        old_type = old_type_func(non_callable_input_types)
        return _nullify_vector_type(old_type)

    return new_callable_nullable_awl_type


def adjust_input_types_for_nullability(
    input_types: InputTypeDict, all_args_nullable: bool
) -> InputTypeDict:
    new_input_type_dict: dict[str, typing.Any] = {}
    for i, key in enumerate(input_types):
        if i == 0 or all_args_nullable:
            if callable(input_types[key]):
                new_input_type_dict[key] = _make_nullable_vector_input_type_callable(
                    typing.cast(TypeCallable, input_types[key])
                )
            else:
                new_input_type_dict[key] = _nullify_vector_type(
                    typing.cast(types.Type, input_types[key])
                )
        else:
            new_input_type_dict[key] = input_types[key]
    return typing.cast(InputTypeDict, new_input_type_dict)


def _handle_arrow_tags(
    old_output_type: ArrowWeaveListType, first_input_type: ArrowWeaveListType
) -> types.Type:
    if isinstance(first_input_type.object_type, tagged_value_type.TaggedValueType):
        return ArrowWeaveListType(
            tagged_value_type.TaggedValueType(
                first_input_type.object_type.tag, old_output_type.object_type
            )
        )
    return old_output_type


def _make_new_vector_output_type_callable(
    old_type_func: TypeCallable, is_null_consuming=False
) -> TypeCallable:
    def new_callable_nullable_awl_type(
        non_callable_input_types: dict[str, types.Type]
    ) -> types.Type:
        first_input_type_name = next(k for k in non_callable_input_types)
        first_input_type = non_callable_input_types[first_input_type_name]
        assert ArrowWeaveListType().assign_type(first_input_type)
        first_input_type = typing.cast(ArrowWeaveListType, first_input_type)

        has_optional_vector_type = types.is_optional(first_input_type.object_type)
        if has_optional_vector_type and not is_null_consuming:
            first_input_type = ArrowWeaveListType(
                types.non_none(first_input_type.object_type)
            )
            non_callable_input_types[first_input_type_name] = first_input_type

        old_type = typing.cast(
            ArrowWeaveListType, old_type_func(non_callable_input_types)
        )

        tag_propagated_output_type = _handle_arrow_tags(old_type, first_input_type)

        if has_optional_vector_type and not is_null_consuming:
            return _nullify_vector_type(tag_propagated_output_type)
        return tag_propagated_output_type

    return new_callable_nullable_awl_type


def adjust_output_type_for_tags_and_nullability(
    output_type: TypeOrCallable,
    is_null_consuming=False,
) -> TypeOrCallable:
    if not callable(output_type):
        new_output_type = typing.cast(TypeCallable, lambda _: output_type)
    else:
        new_output_type = typing.cast(TypeCallable, output_type)
    return _make_new_vector_output_type_callable(
        new_output_type, is_null_consuming=is_null_consuming
    )


def is_null_consuming_arrow_op(input_types: InputTypeDict) -> bool:
    first_input_type_name = next(k for k in input_types)
    first_input_type = input_types[first_input_type_name]
    assert not callable(first_input_type)
    assert ArrowWeaveListType().assign_type(first_input_type)
    return types.is_optional(
        typing.cast(ArrowWeaveListType, first_input_type).object_type
    )


def arrow_op(
    input_type: InputTypeDict,
    output_type: TypeOrCallable,
    refine_output_type=None,
    name=None,
    setter=None,
    render_info=None,
    pure=True,
    all_args_nullable: bool = True,
    plugins=None,
) -> typing.Callable[[typing.Any], op_def.OpDef]:
    """An arrow op is an op that should obey element-based tag-flow map rules. An arrow op must

    1) Have a first arg that is an ArrowWeaveList.
    2) Output an ArrowWeaveList with the same shape as (1)
    3) Each element of the output should represent a mapped transform of the input

    In these cases, element tags from (1) are automatically propagated to each element of (2).
    """

    # TODO(DG): handle reading input and output types from function signature
    is_null_consuming = is_null_consuming_arrow_op(input_type)
    if not is_null_consuming:
        new_input_type = adjust_input_types_for_nullability(
            input_type, all_args_nullable
        )
    else:
        new_input_type = input_type
    new_output_type = adjust_output_type_for_tags_and_nullability(
        output_type, is_null_consuming=is_null_consuming
    )

    return op(
        input_type=new_input_type,
        output_type=new_output_type,
        refine_output_type=refine_output_type,
        name=name,
        setter=setter,
        render_info=render_info,
        pure=pure,
        _op_def_class=op_def.AutoTagHandlingArrowOpDef,
        plugins=plugins,
    )  # type: ignore



import asyncio
import inspect
import textwrap
import time
import traceback
import typing
from typing import Any, Callable, Optional, Union

from rich import print
from rich.console import Console

import weave
from weave.flow import util
from weave.flow.dataset import Dataset
from weave.flow.model import Model, get_infer_method
from weave.flow.obj import Object
from weave.flow.scorer import (
    Scorer,
    auto_summarize,
    get_scorer_attributes,
    transpose,
)
from weave.trace.env import get_weave_parallelism
from weave.trace.errors import OpCallError
from weave.trace.op import Op, as_op, is_op
from weave.trace.vals import WeaveObject
from weave.trace.weave_client import get_ref

console = Console()


INVALID_MODEL_ERROR = (
    "`Evaluation.evaluate` requires a `Model` or `Op` instance as the `model` argument. "
    + "If you are using a function, wrap it with `weave.op` to create an `Op` instance."
)


def async_call(
    func: typing.Union[Callable, Op], *args: Any, **kwargs: Any
) -> typing.Coroutine:
    is_async = False
    if is_op(func):
        func = as_op(func)
        is_async = inspect.iscoroutinefunction(func.resolve_fn)
    else:
        is_async = inspect.iscoroutinefunction(func)
    if is_async:
        return func(*args, **kwargs)  # type: ignore
    return asyncio.to_thread(func, *args, **kwargs)


class EvaluationResults(weave.Object):
    rows: weave.Table


class Evaluation(Object):
    """
    Sets up an evaluation which includes a set of scorers and a dataset.

    Calling evaluation.evaluate(model) will pass in rows from a dataset into a model matching
        the names of the columns of the dataset to the argument names in model.predict.

    Then it will call all of the scorers and save the results in weave.

    If you want to preprocess the rows from the dataset you can pass in a function
    to preprocess_model_input.

    Examples:

    ```python
    # Collect your examples
    examples = [
        {"question": "What is the capital of France?", "expected": "Paris"},
        {"question": "Who wrote 'To Kill a Mockingbird'?", "expected": "Harper Lee"},
        {"question": "What is the square root of 64?", "expected": "8"},
    ]

    # Define any custom scoring function
    @weave.op()
    def match_score1(expected: str, model_output: dict) -> dict:
        # Here is where you'd define the logic to score the model output
        return {'match': expected == model_output['generated_text']}

    @weave.op()
    def function_to_evaluate(question: str):
        # here's where you would add your LLM call and return the output
        return  {'generated_text': 'Paris'}

    # Score your examples using scoring functions
    evaluation = Evaluation(
        dataset=examples, scorers=[match_score1]
    )

    # Start tracking the evaluation
    weave.init('intro-example')
    # Run the evaluation
    asyncio.run(evaluation.evaluate(function_to_evaluate))
    ```
    """

    dataset: Union[Dataset, list]
    scorers: Optional[list[Union[Callable, Op, Scorer]]] = None
    preprocess_model_input: Optional[Callable] = None
    trials: int = 1

    def model_post_init(self, __context: Any) -> None:
        scorers: list[Union[Callable, Scorer, Op]] = []
        for scorer in self.scorers or []:
            if isinstance(scorer, Scorer):
                pass
            elif isinstance(scorer, type):
                raise ValueError(
                    f"Scorer {scorer.__name__} must be an instance, not a class. Did you forget to instantiate?"
                )
            elif callable(scorer) and not is_op(scorer):
                scorer = weave.op()(scorer)
            elif is_op(scorer):
                pass
            else:
                raise ValueError(f"Invalid scorer: {scorer}")
            scorers.append(scorer)
        self.scorers = scorers

        if isinstance(self.dataset, list):
            self.dataset = Dataset(rows=self.dataset)

        if self.name is None and self.dataset.name is not None:
            self.name = self.dataset.name + "-evaluation"  # type: ignore

    @weave.op()
    async def predict_and_score(
        self, model: Union[Callable, Model], example: dict
    ) -> dict:
        if self.preprocess_model_input is None:
            model_input = example
        else:
            model_input = self.preprocess_model_input(example)  # type: ignore

        if callable(model):
            model_predict = model
        else:
            model_predict = get_infer_method(model)

        model_predict_fn_name = (
            as_op(model_predict).name
            if is_op(model_predict)
            else model_predict.__name__
        )

        if is_op(model_predict):
            model_predict = as_op(model_predict)
            predict_signature = model_predict.signature
        else:
            predict_signature = inspect.signature(model_predict)
        model_predict_arg_names = list(predict_signature.parameters.keys())

        if isinstance(model_input, dict):
            model_predict_args = {
                k: v for k, v in model_input.items() if k in model_predict_arg_names
            }
        else:
            if len(model_predict_arg_names) == 1:
                model_predict_args = {model_predict_arg_names[0]: model_input}
            else:
                raise ValueError(
                    f"{model_predict} expects arguments: {model_predict_arg_names}, provide a preprocess_model_input function that returns a dict with those keys."
                )
        try:
            model_start_time = time.time()
            model_output = await async_call(model_predict, **model_predict_args)
        except OpCallError as e:
            dataset_column_names = list(example.keys())
            dataset_column_names_str = ", ".join(dataset_column_names[:3])
            if len(dataset_column_names) > 3:
                dataset_column_names_str += ", ..."
            required_arg_names = [
                param.name
                for param in predict_signature.parameters.values()
                if param.default == inspect.Parameter.empty
            ]

            message = textwrap.dedent(
                f"""
                Call error: {e}

                Options for resolving:
                a. change {model_predict_fn_name} argument names to match a subset of dataset column names: {dataset_column_names_str}
                b. change dataset column names to match expected {model_predict_fn_name} argument names: {required_arg_names}
                c. construct Evaluation with a preprocess_model_input function that accepts a dataset example and returns a dict with keys expected by {model_predict_fn_name}
                """
            )
            raise OpCallError(message)
        except Exception as e:
            print("model_output failed")
            traceback.print_exc()
            model_output = None
        model_latency = time.time() - model_start_time

        scores = {}
        scorers = typing.cast(list[Union[Op, Scorer]], self.scorers or [])
        for scorer in scorers:
            scorer_name, score_fn, _ = get_scorer_attributes(scorer)
            if is_op(score_fn):
                score_fn = as_op(score_fn)
                score_signature = score_fn.signature
            else:
                score_signature = inspect.signature(score_fn)
            score_arg_names = list(score_signature.parameters.keys())

            if "model_output" not in score_arg_names:
                raise OpCallError(
                    f"Scorer {scorer_name} must have a 'model_output' argument, to receive the output of the model function."
                )

            if isinstance(example, dict):
                score_args = {k: v for k, v in example.items() if k in score_arg_names}
            else:
                if len(score_arg_names) == 2:
                    score_args = {score_arg_names[0]: example}
                else:
                    raise ValueError(
                        f"{score_fn} expects arguments: {score_arg_names}, provide a preprocess_model_input function that returns a dict with those keys."
                    )
            score_args["model_output"] = model_output

            try:
                result = await async_call(score_fn, **score_args)
            except OpCallError as e:
                dataset_column_names = list(example.keys())
                dataset_column_names_str = ", ".join(dataset_column_names[:3])
                if len(dataset_column_names) > 3:
                    dataset_column_names_str += ", ..."
                required_arg_names = [
                    param.name
                    for param in score_signature.parameters.values()
                    if param.default == inspect.Parameter.empty
                ]
                required_arg_names.remove("model_output")

                message = textwrap.dedent(
                    f"""
                    Call error: {e}

                    Options for resolving:
                    a. change {scorer_name} argument names to match a subset of dataset column names ({dataset_column_names_str})
                    b. change dataset column names to match expected {scorer_name} argument names: {required_arg_names}
                    """
                )
                raise OpCallError(message)
            scores[scorer_name] = result

        return {
            "model_output": model_output,
            "scores": scores,
            "model_latency": model_latency,
        }

    @weave.op()
    async def summarize(self, eval_table: EvaluationResults) -> dict:
        eval_table_rows = list(eval_table.rows)
        cols = transpose(eval_table_rows)
        summary = {}

        for name, vals in cols.items():
            if name == "scores":
                scorers = self.scorers or []
                for scorer in scorers:
                    scorer_name, _, summarize_fn = get_scorer_attributes(scorer)
                    scorer_stats = transpose(vals)
                    score_table = scorer_stats[scorer_name]
                    scored = summarize_fn(score_table)
                    summary[scorer_name] = scored
            else:
                model_output_summary = auto_summarize(vals)
                if model_output_summary:
                    summary[name] = model_output_summary

        return summary

    @weave.op()
    async def evaluate(self, model: Union[Callable, Model]) -> dict:
        if not is_valid_model(model):
            raise ValueError(INVALID_MODEL_ERROR)
        eval_rows = []

        start_time = time.time()

        async def eval_example(example: dict) -> dict:
            try:
                eval_row = await self.predict_and_score(model, example)
            except OpCallError as e:
                raise e
            except Exception as e:
                print("Predict and score failed")
                traceback.print_exc()
                return {"model_output": None, "scores": {}}
            return eval_row

        n_complete = 0
        # with console.status("Evaluating...") as status:
        dataset = typing.cast(Dataset, self.dataset)
        _rows = dataset.rows
        trial_rows = list(_rows) * self.trials
        async for example, eval_row in util.async_foreach(
            trial_rows, eval_example, get_weave_parallelism()
        ):
            n_complete += 1
            duration = time.time() - start_time
            print(f"Evaluated {n_complete} of {len(trial_rows)} examples")
            # status.update(
            #     f"Evaluating... {duration:.2f}s [{n_complete} / {len(self.dataset.rows)} complete]"  # type:ignore
            # )
            if eval_row is None:
                eval_row = {"model_output": None, "scores": {}}
            else:
                eval_row["scores"] = eval_row.get("scores", {})
            for scorer in self.scorers or []:
                scorer_name, _, _ = get_scorer_attributes(scorer)
                if scorer_name not in eval_row["scores"]:
                    eval_row["scores"][scorer_name] = {}
            eval_rows.append(eval_row)

        # The need for this pattern is quite unfortunate and highlights a gap in our
        # data model. As a user, I just want to pass a list of data `eval_rows` to
        # summarize. Under the hood, Weave should choose the appropriate storage
        # format (in this case `Table`) and serialize it that way. Right now, it is
        # just a huge list of dicts. The fact that "as a user" I need to construct
        # `weave.Table` at all is a leaky abstraction. Moreover, the need to
        # construct `EvaluationResults` just so that tracing and the UI works is
        # also bad. In the near-term, this will at least solve the problem of
        # breaking summarization with big datasets, but this is not the correct
        # long-term solution.
        eval_results = EvaluationResults(rows=weave.Table(eval_rows))
        summary = await self.summarize(eval_results)

        print("Evaluation summary", summary)

        return summary


def evaluate(
    dataset: Union[Dataset, list],
    model: Union[Callable, Model],
    scores: Optional[list[Union[Callable, Scorer]]] = None,
    preprocess_model_input: Optional[Callable] = None,
) -> dict:
    eval = Evaluation(
        dataset=dataset, scorers=scores, preprocess_model_input=preprocess_model_input
    )
    return asyncio.run(eval.evaluate(model))


def is_valid_model(model: Any) -> bool:
    return (
        # Model instances are supported
        isinstance(model, Model)
        # Ops are supported
        or is_op(model)
        # Saved Models (Objects with predict) are supported
        or (
            get_ref(model) is not None
            and isinstance(model, WeaveObject)
            and hasattr(model, "predict")
            and is_op(model.predict)
        )
    )



import typing

from weave.legacy.weave import weave_types as types
from weave.legacy.weave.decorator_type import type as weave_type
from weave.legacy.weave.partial_object import (
    PartialObject,
    PartialObjectTypeGeneratorType,
)

"""
This file contains all the "W&B Domain Types". Each domain type should
correspond to exactly 1 type/interface in the WB GQL schema. Not all types are
represented as we only have implemented the subset that is required for Weave1
compatibility. Each type follows the same pattern for declaration. Consider
User:

@weave_type("user", True)                       # Sets the Weave type to be a plain string called "user"
class User(GQLTypeMixin):                       # Creates the Weave type in Weave1 and adds the GQLTypeMixin
    REQUIRED_FRAGMENT = "id name"               # Declares a minimum GQL fragment that should be used whenever loading this type from GQL

UserType = User.WeaveType()  # type: ignore     # Exports a casted singleton type for easier construction in domain ops.
UserType = typing.cast(types.Type, UserType) .  # Note: the type: ignore and cast could be removed if we fix the .pyi file for decorator_type

There are two main concepts to grasp:
    1. The `REQUIRED_FRAGMENT` is used to specify the minimum GQL fragment that
       should be used when loading this type from GQL. In most cases
        this is just the ID and name, but in some cases we need more
        information. For example, when loading a Run, we also load the Project
        (and extension, the project's Entity).
    2. The data is currently stored in `User.gql` - which gets serialized as an
       untyped string, but can be accessed like a dictionary. See
       `UntypedOpaqueDict` for details.

TODO: Decide if the nested fragments are really necessary. If not, it will
require changing a few of the ops to fetch such info as well
      as updating the mocked tests.
"""


def gql_type(
    name: str,
) -> typing.Callable[[typing.Type["GQLBase"]], typing.Type[PartialObject]]:
    """Decorator that emits a Weave Type for the decorated GQL instance type. Classes decorated
    with this decorator also emit a keyless weavetype. E.g.,

    @gql_type("project")
    class Project(GQLBase):
        ...

    will emit a Weave Type for projectType(). Instances of projectType() have a method called
    with_keys() that allows them to generate instances of projectTypeWithKeys({...}).
    """

    def _gql_weave_type(
        _instance_class: typing.Type["GQLBase"],
    ) -> typing.Type[PartialObject]:
        decorator = weave_type(
            name,
            True,
            None,
            [PartialObjectTypeGeneratorType],
        )
        return decorator(_instance_class)

    return _gql_weave_type


class GQLBase(PartialObject):
    REQUIRED_FRAGMENT: str


@gql_type("org")
class Org(GQLBase):
    REQUIRED_FRAGMENT = """
        id
        name
    """


OrgType = Org.WeaveType()  # type: ignore
OrgType = typing.cast(PartialObjectTypeGeneratorType, OrgType)


@gql_type("entity")
class Entity(GQLBase):
    REQUIRED_FRAGMENT = """
        id
        name
    """


EntityType = Entity.WeaveType()  # type: ignore
EntityType = typing.cast(PartialObjectTypeGeneratorType, EntityType)


@gql_type("user")
class User(GQLBase):
    REQUIRED_FRAGMENT = """
        id
        name
    """


UserType = User.WeaveType()  # type: ignore
UserType = typing.cast(PartialObjectTypeGeneratorType, UserType)


@gql_type("project")
class Project(GQLBase):
    REQUIRED_FRAGMENT = """
        id
        name
    """


ProjectType = Project.WeaveType()  # type: ignore
ProjectType = typing.cast(PartialObjectTypeGeneratorType, ProjectType)


@gql_type("run")
class Run(GQLBase):
    REQUIRED_FRAGMENT = """
        id
        name
    """


RunType = Run.WeaveType()  # type: ignore
RunType = typing.cast(PartialObjectTypeGeneratorType, RunType)


@gql_type("artifactType")
class ArtifactType(GQLBase):
    REQUIRED_FRAGMENT = """
        id
        name
    """


ArtifactTypeType = ArtifactType.WeaveType()  # type: ignore
ArtifactTypeType = typing.cast(PartialObjectTypeGeneratorType, ArtifactTypeType)


@gql_type("artifact")  # Name and Class mismatch intention due to weave0
class ArtifactCollection(GQLBase):
    REQUIRED_FRAGMENT = """
        id
        name
    """


ArtifactCollectionType = ArtifactCollection.WeaveType()  # type: ignore
ArtifactCollectionType = typing.cast(
    PartialObjectTypeGeneratorType, ArtifactCollectionType
)


@gql_type("artifactVersion")
class ArtifactVersion(GQLBase):
    REQUIRED_FRAGMENT = """
        id
    """


ArtifactVersionType = ArtifactVersion.WeaveType()  # type: ignore
ArtifactVersionType = typing.cast(PartialObjectTypeGeneratorType, ArtifactVersionType)


@gql_type("artifactMembership")  # Name and Class mismatch intention due to weave0
class ArtifactCollectionMembership(GQLBase):
    REQUIRED_FRAGMENT = """
        id
    """


ArtifactCollectionMembershipType = ArtifactCollectionMembership.WeaveType()  # type: ignore
ArtifactCollectionMembershipType = typing.cast(
    PartialObjectTypeGeneratorType, ArtifactCollectionMembershipType
)


@gql_type("artifactAlias")
class ArtifactAlias(GQLBase):
    REQUIRED_FRAGMENT = """
        id
    """


ArtifactAliasType = ArtifactAlias.WeaveType()  # type: ignore
ArtifactAliasType = typing.cast(PartialObjectTypeGeneratorType, ArtifactAliasType)


@gql_type("report")
class Report(GQLBase):
    REQUIRED_FRAGMENT = """
        id
    """


ReportType = Report.WeaveType()  # type: ignore
ReportType = typing.cast(PartialObjectTypeGeneratorType, ReportType)


@gql_type("runQueue")
class RunQueue(GQLBase):
    REQUIRED_FRAGMENT = """
        id
    """


RunQueueType = RunQueue.WeaveType()  # type: ignore
RunQueueType = typing.cast(PartialObjectTypeGeneratorType, RunQueueType)


# Simple types (maybe should be put into primitives?)


@weave_type("link", __is_simple=True)  # type: ignore
class Link:
    name: str
    url: str


LinkType = Link.WeaveType()  # type: ignore
LinkType = typing.cast(types.Type, LinkType)



import pytest

import weave

from ...trace_server import trace_server_interface as tsi

# This file tests the different argument variations that can be passed to an op.
#
#
# There are 24 (2 * 2 * 3 * 2) distinct arg shapes, each with a number of variants
#
#                        Astrisk in the table means that the argument is named
#                        we use `*` to indicate the start of keyword-only arguments
#                        rather than using default values ---\
#                                                            |
#                                                            |
# | Example Stub                    | PosArg  | VarArg   | KwArg   | VarKwArg | Variants |
# |---------------------------------|---------|----------|---------|----------|----------|
# | fn()                            |   NO    |    NO    |  NO     |   NO     |    1     |
# | fn(**kwargs)                    |   NO    |    NO    |  NO     |   YES    |    2     |
# | fn(y=0)                         |   NO    |    NO    |  YES    |   NO     |    2     |
# | fn(*, y)                        |   NO    |    NO    |  YES*   |   NO     |    1     |
# | fn(y=0, **kwargs)               |   NO    |    NO    |  YES    |   YES    |    4 + 2 |
# | fn(*, y, **kwargs)              |   NO    |    NO    |  YES*   |   YES    |    2     |
# | fn(*args)                       |   NO    |    YES   |  NO     |   NO     |    2     |
# | fn(*args, **kwargs)             |   NO    |    YES   |  NO     |   YES    |    4     |
# | fn(*args, y=0)                  |   NO    |    YES   |  YES    |   NO     |    4     |
# | fn(*args, y)                    |   NO    |    YES   |  YES*   |   NO     |    2     |
# | fn(*args, y=0, **kwargs)        |   NO    |    YES   |  YES    |   YES    |    8     |
# | fn(*args, y, **kwargs)          |   NO    |    YES   |  YES*   |   YES    |    4     |
# | fn(x)                           |   YES   |    NO    |  NO     |   NO     |    2     |
# | fn(x, **kwargs)                 |   YES   |    NO    |  NO     |   YES    |    4     |
# | fn(x, y=0)                      |   YES   |    NO    |  YES    |   NO     |    4 + 1 |
# | fn(x, *, y)                     |   YES   |    NO    |  YES*   |   NO     |    2     |
# | fn(x, y=0, **kwargs)            |   YES   |    NO    |  YES    |   YES    |    8     |
# | fn(x, *, y, **kwargs)           |   YES   |    NO    |  YES*   |   YES    |    4     |
# | fn(x, *args)                    |   YES   |    YES   |  NO     |   NO     |    4 - 1 |
# | fn(x, *args, **kwargs)          |   YES   |    YES   |  NO     |   YES    |    8 - 2 |
# | fn(x, *args, y=0)               |   YES   |    YES   |  YES    |   NO     |    8 - 2 |
# | fn(x, *args, y)                 |   YES   |    YES   |  YES*   |   NO     |    4 - 1 |
# | fn(x, *args, y=0, **kwargs)     |   YES   |    YES   |  YES    |   YES    |    16 - 4|
# | fn(x, *args, y, **kwargs)       |   YES   |    YES   |  YES*   |   YES    |    8 - 2 |
#
# Within each stub, we will have the following variants:
# * PosArg: If YES, 2x: both with and without keyword
# * VarArg: If YES, 2x: both with and without values
# * KwArg:
#           If YES, 2x: both with and without values.
#           If YES*, 1x
# * VarKwArg: If YES, 2x: both with and without values
#
# Some stubs have a few patterns added or removed to account for python's rules. The +/- indicates the number of patterns
# added or removed.


@pytest.mark.parametrize(
    "fn, arg_variations",
    [
        # | fn()                            |   NO    |    NO    |  NO     |   NO     |    1     |
        (
            (lambda: {}),
            [
                ((), {}),
            ],
        ),
        # | fn(**kwargs)                    |   NO    |    NO    |  NO     |   YES    |    2     |
        (
            (lambda **kwargs: {"kwargs": kwargs}),
            [
                ((), {}),
                ((), {"a": 1}),
            ],
        ),
        # | fn(y=0)                         |   NO    |    NO    |  YES    |   NO     |    2     |
        (
            (lambda y=0: {"y": y}),
            [
                ((), {}),
                ((), {"y": 1}),
            ],
        ),
        # | fn(*, y)                        |   NO    |    NO    |  YES*   |   NO     |    1     |
        (
            (lambda *, y=0: {"y": y}),
            [
                ((), {"y": 1}),
            ],
        ),
        # | fn(y=0, **kwargs)               |   NO    |    NO    |  YES    |   YES    |    4 + 2 |
        (
            (lambda y=0, **kwargs: {"y": y, "kwargs": kwargs}),
            [
                ((), {}),
                ((1,), {}),  # Extra pattern possible with this combination
                ((), {"y": 1}),
                ((), {"a": 1}),
                ((1,), {"a": 2}),  # Extra pattern possible with this combination
                ((), {"y": 1, "a": 2}),
            ],
        ),
        # | fn(*, y, **kwargs)              |   NO    |    NO    |  YES*   |   YES    |    2     |
        (
            (lambda *, y, **kwargs: {"y": y, "kwargs": kwargs}),
            [
                ((), {"y": 1}),
                ((), {"y": 1, "a": 2}),
            ],
        ),
        # | fn(*args)                       |   NO    |    YES   |  NO     |   NO     |    2     |
        (
            (lambda *args: {"args": list(args)}),
            [
                ((), {}),
                ((1,), {}),
            ],
        ),
        # | fn(*args, **kwargs)             |   NO    |    YES   |  NO     |   YES    |    4     |
        (
            (lambda *args, **kwargs: {"args": list(args), "kwargs": kwargs}),
            [
                ((), {}),
                ((1,), {}),
                ((), {"a": 1}),
                ((1,), {"a": 2}),
            ],
        ),
        # | fn(*args, y=0)                  |   NO    |    YES   |  YES    |   NO     |    4     |
        (
            (lambda *args, y=0: {"args": list(args), "y": y}),
            [
                ((), {}),
                ((1,), {}),
                ((), {"y": 1}),
                ((1,), {"y": 2}),
            ],
        ),
        # | fn(*args, y)                    |   NO    |    YES   |  YES*   |   NO     |    2     |
        (
            (lambda *args, y: {"args": list(args), "y": y}),
            [
                ((), {"y": 1}),
                ((1,), {"y": 2}),
            ],
        ),
        # | fn(*args, y=0, **kwargs)        |   NO    |    YES   |  YES    |   YES    |    8     |
        (
            (
                lambda *args, y=0, **kwargs: {
                    "args": list(args),
                    "y": y,
                    "kwargs": kwargs,
                }
            ),
            [
                ((), {}),
                ((1,), {}),
                ((), {"y": 1}),
                ((), {"a": 1}),
                ((1,), {"y": 2}),
                ((), {"y": 1, "a": 2}),
                ((1,), {"a": 2}),
                ((1,), {"y": 2, "a": 2}),
            ],
        ),
        # | fn(*args, y, **kwargs)          |   NO    |    YES   |  YES*   |   YES    |    4     |
        (
            (
                lambda *args, y, **kwargs: {
                    "args": list(args),
                    "y": y,
                    "kwargs": kwargs,
                }
            ),
            [
                ((), {"y": 1}),
                ((1,), {"y": 2}),
                ((), {"y": 1, "a": 2}),
                ((1,), {"y": 2, "a": 3}),
            ],
        ),
        # | fn(x)                           |   YES   |    NO    |  NO     |   NO     |    2     |
        (
            (lambda x: {"x": x}),
            [
                ((), {"x": 1}),
                ((1,), {}),
            ],
        ),
        # | fn(x, **kwargs)                 |   YES   |    NO    |  NO     |   YES    |    4     |
        (
            (lambda x, **kwargs: {"x": x, "kwargs": kwargs}),
            [
                ((), {"x": 1}),
                ((1,), {}),
                ((), {"x": 1, "a": 2}),
                ((1,), {"a": 2}),
            ],
        ),
        # | fn(x, y=0)                      |   YES   |    NO    |  YES    |   NO     |    4 + 1 |
        (
            (lambda x, y=0: {"x": x, "y": y}),
            [
                ((), {"x": 1}),
                ((1,), {}),
                ((1, 2), {}),  # Extra pattern possible with this combination
                ((), {"x": 1, "y": 2}),
                ((1,), {"y": 2}),
            ],
        ),
        # | fn(x, *, y)                     |   YES   |    NO    |  YES*   |   NO     |    2     |
        (
            (lambda x, *, y: {"x": x, "y": y}),
            [
                ((), {"x": 1, "y": 2}),
                ((1,), {"y": 2}),
            ],
        ),
        # | fn(x, y=0, **kwargs)            |   YES   |    NO    |  YES    |   YES    |    8     |
        (
            (lambda x, y=0, **kwargs: {"x": x, "y": y, "kwargs": kwargs}),
            [
                ((), {"x": 1}),
                ((1,), {}),
                ((), {"x": 1, "y": 2}),
                ((1,), {"y": 2}),
                ((), {"x": 1, "a": 2}),
                ((1,), {"a": 2}),
                ((), {"x": 1, "y": 2, "a": 3}),
                ((1,), {"y": 2, "a": 3}),
            ],
        ),
        # | fn(x, *, y, **kwargs)           |   YES   |    NO    |  YES*   |   YES    |    4     |
        (
            (lambda x, *, y, **kwargs: {"x": x, "y": y, "kwargs": kwargs}),
            [
                ((), {"x": 1, "y": 2}),
                ((1,), {"y": 2}),
                ((), {"x": 1, "y": 2, "a": 3}),
                ((1,), {"y": 2, "a": 3}),
            ],
        ),
        # | fn(x, *args)                    |   YES   |    YES   |  NO     |   NO     |    4 - 1 |
        (
            (lambda x, *args: {"x": x, "args": list(args)}),
            [
                ((), {"x": 1}),
                ((1,), {}),
                # ((1,), {"x": 2}), # Invalid
                (
                    (
                        1,
                        2,
                    ),
                    {},
                ),
            ],
        ),
        # | fn(x, *args, **kwargs)          |   YES   |    YES   |  NO     |   YES    |    8 - 2 |
        (
            (lambda x, *args, **kwargs: {"x": x, "args": list(args), "kwargs": kwargs}),
            [
                ((), {"x": 1}),
                ((1,), {}),
                # ((1,), {"x": 2}), # Invalid
                (
                    (
                        1,
                        2,
                    ),
                    {},
                ),
                ((), {"x": 1, "a": 2}),
                ((1,), {"a": 2}),
                # ((1,), {"x": 2, "a": 3}), # Invalid
                ((1, 2), {"a": 3}),
            ],
        ),
        # | fn(x, *args, y=0)               |   YES   |    YES   |  YES    |   NO     |    8 - 2 |
        (
            (lambda x, *args, y=0: {"x": x, "args": list(args), "y": y}),
            [
                ((), {"x": 1}),
                ((1,), {}),
                # ((1,), {"x": 2}), # Invalid
                (
                    (
                        1,
                        2,
                    ),
                    {},
                ),
                ((), {"x": 1, "y": 2}),
                ((1,), {"y": 2}),
                # ((1,), {"x": 2, "y": 3}), # Invalid
                (
                    (
                        1,
                        2,
                    ),
                    {"y": 3},
                ),
            ],
        ),
        # | fn(x, *args, y)                 |   YES   |    YES   |  YES*   |   NO     |    4 - 1 |
        (
            (lambda x, *args, y: {"x": x, "args": list(args), "y": y}),
            [
                ((), {"x": 1, "y": 2}),
                ((1,), {"y": 2}),
                # ((1,), {"x": 2, "y": 3}), # Invalid
                (
                    (
                        1,
                        2,
                    ),
                    {"y": 3},
                ),
            ],
        ),
        # | fn(x, *args, y=0, **kwargs)     |   YES   |    YES   |  YES    |   YES    |    16 - 4|
        (
            (
                lambda x, *args, y=0, **kwargs: {
                    "x": x,
                    "args": list(args),
                    "y": y,
                    "kwargs": kwargs,
                }
            ),
            [
                ((), {"x": 1}),
                ((1,), {}),
                # ((1,), {"x": 2}), # Invalid
                (
                    (
                        1,
                        2,
                    ),
                    {},
                ),
                ((), {"x": 1, "y": 2}),
                ((1,), {"y": 2}),
                # ((1,), {"x": 2, "y": 3}), # Invalid
                (
                    (
                        1,
                        2,
                    ),
                    {"y": 3},
                ),
                ((), {"x": 1, "a": 2}),
                ((1,), {"a": 2}),
                # ((1,), {"x": 2, "a": 3}), # Invalid
                (
                    (
                        1,
                        2,
                    ),
                    {"a": 3},
                ),
                ((), {"x": 1, "y": 2, "a": 3}),
                ((1,), {"y": 2, "a": 3}),
                # ((1,), {"x": 2, "y": 3, "a": 4}), # Invalid
                (
                    (
                        1,
                        2,
                    ),
                    {"y": 3, "a": 4},
                ),
            ],
        ),
        # | fn(x, *args, y, **kwargs)       |   YES   |    YES   |  YES*   |   YES    |    8 - 2|
        (
            (
                lambda x, *args, y, **kwargs: {
                    "x": x,
                    "args": list(args),
                    "y": y,
                    "kwargs": kwargs,
                }
            ),
            [
                ((), {"x": 1, "y": 2}),
                ((1,), {"y": 2}),
                # ((1,), {"x": 2, "y": 3}), # Invalid
                (
                    (
                        1,
                        2,
                    ),
                    {"y": 3},
                ),
                ((), {"x": 1, "y": 2, "a": 3}),
                ((1,), {"y": 2, "a": 3}),
                # ((1,), {"x": 2, "y": 3, "a": 4}), # Invalid
                (
                    (
                        1,
                        2,
                    ),
                    {"y": 3, "a": 4},
                ),
            ],
        ),
    ],
)
def test_general_arg_variations(client, fn, arg_variations):
    """In this test, `fn` is expect to return the value of the inputs"""
    wrapped_fn = weave.op()(fn)
    for ndx, variation in enumerate(arg_variations):
        args = variation[0]
        kwargs = variation[1]

        fn_res = fn(*args, **kwargs)
        wrapped_res = wrapped_fn(*args, **kwargs)

        assert fn_res == wrapped_res

        res = client.server.calls_query(
            tsi.CallsQueryReq(
                project_id=client._project_id(),
                filter=tsi.CallsFilter(op_names=[wrapped_fn.ref.uri()]),
            )
        )

        assert len(res.calls) == ndx + 1
        assert res.calls[ndx].inputs == fn_res
        assert res.calls[ndx].output == fn_res


# Below are specific tests for each of the 24 stubs
# While the above test is more general and covers all possible variations, the below tests are more specific and
# a bit easier to read and understand (technically, the above test is more comprehensive and should be enough)


def test_no_args(client):
    @weave.op
    def my_op() -> int:
        return 1

    my_op()

    res = client.server.calls_query(
        tsi.CallsQueryReq(
            project_id=client._project_id(),
        )
    )

    assert res.calls[0].op_name == my_op.ref.uri()
    assert res.calls[0].inputs == {}


def test_args_empty(client):
    @weave.op()
    def my_op() -> int:
        return 1

    my_op()

    res = client.server.calls_query(
        tsi.CallsQueryReq(
            project_id=client._project_id(),
        )
    )

    assert res.calls[0].op_name == my_op.ref.uri()
    assert res.calls[0].inputs == {}


def test_args_concrete(client):
    @weave.op()
    def my_op(val):
        return [val]

    my_op(1)

    res = client.server.calls_query(
        tsi.CallsQueryReq(
            project_id=client._project_id(),
        )
    )

    assert res.calls[0].op_name == my_op.ref.uri()
    assert res.calls[0].inputs == {"val": 1}
    assert res.calls[0].output == [1]


def test_args_concrete_splat(client):
    @weave.op()
    def my_op(val, *args):
        return [val, args]

    my_op(1)
    my_op(1, 2, 3)

    res = client.server.calls_query(
        tsi.CallsQueryReq(
            project_id=client._project_id(),
        )
    )

    assert res.calls[0].op_name == my_op.ref.uri()
    assert res.calls[0].inputs == {"val": 1, "args": []}
    assert res.calls[0].output == [1, []]
    assert res.calls[1].op_name == my_op.ref.uri()
    assert res.calls[1].inputs == {"val": 1, "args": [2, 3]}
    assert res.calls[1].output == [1, [2, 3]]


def test_args_concrete_splats(client):
    @weave.op()
    def my_op(val, *args, **kwargs):
        return [val, args, kwargs]

    my_op(1)
    my_op(1, 2, 3)
    my_op(1, a=2, b=3)
    my_op(1, 2, 3, a=4, b=5)

    res = client.server.calls_query(
        tsi.CallsQueryReq(
            project_id=client._project_id(),
        )
    )

    assert res.calls[0].op_name == my_op.ref.uri()
    assert res.calls[0].inputs == {"val": 1, "args": [], "kwargs": {}}
    assert res.calls[0].output == [1, [], {}]
    assert res.calls[1].op_name == my_op.ref.uri()
    assert res.calls[1].inputs == {"val": 1, "args": [2, 3], "kwargs": {}}
    assert res.calls[1].output == [1, [2, 3], {}]
    assert res.calls[2].op_name == my_op.ref.uri()
    assert res.calls[2].inputs == {"val": 1, "args": [], "kwargs": {"a": 2, "b": 3}}
    assert res.calls[2].output == [1, [], {"a": 2, "b": 3}]
    assert res.calls[3].op_name == my_op.ref.uri()
    assert res.calls[3].inputs == {"val": 1, "args": [2, 3], "kwargs": {"a": 4, "b": 5}}
    assert res.calls[3].output == [1, [2, 3], {"a": 4, "b": 5}]


def test_args_concrete_splat_concrete(client):
    @weave.op()
    def my_op(val, *args, a=0):
        return [val, args, a]

    my_op(1)
    my_op(1, a=2)
    my_op(1, 2, 3, a=4)

    res = client.server.calls_query(
        tsi.CallsQueryReq(
            project_id=client._project_id(),
        )
    )

    assert res.calls[0].op_name == my_op.ref.uri()
    assert res.calls[0].inputs == {"val": 1, "args": [], "a": 0}
    assert res.calls[0].output == [1, [], 0]
    assert res.calls[1].op_name == my_op.ref.uri()
    assert res.calls[1].inputs == {"val": 1, "args": [], "a": 2}
    assert res.calls[1].output == [1, [], 2]
    assert res.calls[2].op_name == my_op.ref.uri()
    assert res.calls[2].inputs == {"val": 1, "args": [2, 3], "a": 4}
    assert res.calls[2].output == [1, [2, 3], 4]


def test_args_concrete_splat_concrete_splat(client):
    @weave.op()
    def my_op(val, *args, a=0, **kwargs):
        return [val, args, a, kwargs]

    my_op(1)
    my_op(1, a=2)
    my_op(1, 2, 3, a=4)
    my_op(1, 2, 3, a=4, b=5)

    res = client.server.calls_query(
        tsi.CallsQueryReq(
            project_id=client._project_id(),
        )
    )

    assert res.calls[0].op_name == my_op.ref.uri()
    assert res.calls[0].inputs == {"val": 1, "args": [], "a": 0, "kwargs": {}}
    assert res.calls[0].output == [1, [], 0, {}]
    assert res.calls[1].op_name == my_op.ref.uri()
    assert res.calls[1].inputs == {"val": 1, "args": [], "a": 2, "kwargs": {}}
    assert res.calls[1].output == [1, [], 2, {}]
    assert res.calls[2].op_name == my_op.ref.uri()
    assert res.calls[2].inputs == {"val": 1, "args": [2, 3], "a": 4, "kwargs": {}}
    assert res.calls[2].output == [1, [2, 3], 4, {}]
    assert res.calls[3].op_name == my_op.ref.uri()
    assert res.calls[3].inputs == {"val": 1, "args": [2, 3], "a": 4, "kwargs": {"b": 5}}
    assert res.calls[3].output == [1, [2, 3], 4, {"b": 5}]



import weave
from weave.trace.vals import WeaveObject


def test_weaveobject_properties():
    class A:
        @property
        def x(self):
            return 1

    to = WeaveObject(A(), None, None, None)
    assert to.x == 1


def test_weaveobject_access_after_init_termination(client):
    my_obj = None

    class MyObj(weave.Object):
        val: int

    @weave.op()
    def my_op(obj: MyObj) -> None:
        nonlocal my_obj
        my_obj = obj

    my_op(MyObj(val=1))

    assert my_obj.val == 1

    # Here we explicitly close the client in order to
    # simulate a situation where the client is closed
    # but a reference to a trace object still exists.

    weave.trace.client_context.weave_client.set_weave_client_global(None)

    assert my_obj.val == 1



import pytest

import weave
from weave.legacy.weave import storage, weave_internal
from weave.legacy.weave.panels import panel_plot

from ...legacy.weave.panels.panel_group import Group
from ...legacy.weave.panels.panel_slider import Slider


def test_panel_id():
    panel = Group(items={})
    assert panel.id == "Group"
    assert panel.to_json()["id"] == "Group"


@pytest.mark.skip()
# Doesn't work because we process variables "inside-out"
def test_simple_nested_1():
    panel = Group(
        vars={"a": weave_internal.make_const_node(weave.types.Int(), 5)},
        items={
            "item": Group(
                vars={"b": weave_internal.make_const_node(weave.types.Int(), 9)},
                items={"item_inner": lambda a, b: a + b},
            )
        },
    )
    panel._normalize()
    assert str(panel.items["item"].items["item_inner"]) == "add(a, b)"


def test_simple_nested_outer_lambda():
    panel = Group(
        vars={"a": weave_internal.make_const_node(weave.types.Int(), 5)},
        items={
            "item": lambda a: Group(
                vars={"b": weave_internal.make_const_node(weave.types.Int(), 9)},
                items={"item_inner": lambda b: a + b},
            )
        },
    )
    panel._normalize()
    assert str(panel.config.items["item"].config.items["item_inner"]) == "add(a, b)"


@pytest.mark.skip()
def test_nested():
    # Doesn't quite work, we get the wrong type for
    # a.items['a2'] (we get Panel, actually want Panel[number] I think?)
    panel = Group(
        items={
            "a": Group(items={"a1": 5, "a2": lambda a1: a1 + 5}),
            "b": Group(
                items={"b1": lambda a: a.items["a1"], "b2": lambda a: a.items["a2"]}
            ),
        }
    )
    panel.normalize()
    assert 1 == 2


@pytest.mark.skip()
def test_synced():
    panel = Group(
        vars={"num": 0},
        items={
            "a": lambda num: Slider2(num),
            "b": lambda num: Slider2(num),
            "c": lambda num: num,
        },
    )
    print(panel.to_json())
    assert 1 == 2


# def test_select_row():
#     panel = Group(
#         items={"step": Slider2(), "table": panellambda my_slider: my_slider.value}
# )


def test_object_picker_choice_type():
    ints = weave.save([1, 2, 3], name="my-ints")
    panel = weave.legacy.weave.panels.ObjectPicker(ints)
    panel_node = weave_internal.make_var_node(weave.type_of(panel), "panel")
    choice = panel_node.config.choice
    assert choice.type == weave.types.Function({}, weave.types.Int())


def test_facet_selected():
    data = weave.save(
        [{"guess": "dog", "truth": "cat"}, {"guess": "dog", "truth": "dog"}]
    )
    facet = weave.legacy.weave.panels.Group(
        equalSize=True,
        items={
            "confusion": weave.legacy.weave.panels.Facet(
                data,
                x=lambda row: row["guess"],
                y=lambda row: row["truth"],
                select=lambda row: weave.legacy.weave.panels.Group(
                    layered=True,
                    items={
                        "color": weave.legacy.weave.panels.Color(row.count() / 50),
                        "count": row.count(),
                    },
                ),
            ),
            "selected": lambda confusion: confusion.selected(),
        },
    )
    assert facet.config.items["selected"].type == data.type


def test_board():
    # Just make sure it runs for now.
    weave.legacy.weave.panels.Board(
        {
            "nums": weave.legacy.weave.ops.range(0, 3, 1),
        },
        [weave.legacy.weave.panels.BoardPanel(id="panel0", panel=lambda nums: nums)],
    )


def test_plot_constants_assign():
    assert panel_plot.PlotConstants.WeaveType().assign_type(
        weave.type_of(panel_plot.PlotConstants())
    )


def test_plot_assign():
    assert weave.legacy.weave.panels.Plot.WeaveType().assign_type(
        weave.type_of(weave.legacy.weave.panels.Plot([{"a": 5}]))
    )



import dataclasses
import inspect
import logging
import operator
import typing
from typing import Any, Generator, Iterator, Literal, Optional, SupportsIndex, Union

from pydantic import BaseModel
from pydantic import v1 as pydantic_v1

from weave.trace import box
from weave.trace.client_context.weave_client import get_weave_client
from weave.trace.context import get_raise_on_captured_errors
from weave.trace.errors import InternalError
from weave.trace.object_record import ObjectRecord
from weave.trace.op import is_op, maybe_bind_method
from weave.trace.refs import (
    DICT_KEY_EDGE_NAME,
    LIST_INDEX_EDGE_NAME,
    OBJECT_ATTR_EDGE_NAME,
    TABLE_ROW_ID_EDGE_NAME,
    ObjectRef,
    RefWithExtra,
    TableRef,
)
from weave.trace.serialize import from_json
from weave.trace.table import Table
from weave.trace_server.trace_server_interface import (
    ObjReadReq,
    TableQueryReq,
    TableRowFilter,
    TraceServerInterface,
)

logger = logging.getLogger(__name__)


@dataclasses.dataclass
class MutationSetitem:
    path: tuple[str, ...]
    operation: Literal["setitem"]
    args: tuple[str, Any]


@dataclasses.dataclass
class MutationSetattr:
    path: tuple[str, ...]
    operation: Literal["setattr"]
    args: tuple[str, Any]


@dataclasses.dataclass
class MutationAppend:
    path: tuple[str, ...]
    operation: Literal["append"]
    args: tuple[Any]


Mutation = Union[MutationSetattr, MutationSetitem, MutationAppend]
MutationOperation = Union[Literal["setitem"], Literal["setattr"], Literal["append"]]


def make_mutation(
    path: tuple[str, ...], operation: MutationOperation, args: tuple[Any, ...]
) -> Mutation:
    if operation == "setitem":
        if len(args) != 2 or not isinstance(args[0], str):
            raise ValueError("setitem mutation requires 2 args")
        args = typing.cast(tuple[str, Any], args)
        return MutationSetitem(path, operation, args)
    elif operation == "setattr":
        if len(args) != 2 or not isinstance(args[0], str):
            raise ValueError("setattr mutation requires 2 args")
        args = typing.cast(tuple[str, Any], args)
        return MutationSetattr(path, operation, args)
    elif operation == "append":
        if len(args) != 1:
            raise ValueError("append mutation requires 1 arg")
        args = typing.cast(tuple[Any], args)
        return MutationAppend(path, operation, args)
    else:
        raise ValueError(f"Unknown operation: {operation}")


class Traceable:
    ref: Optional[RefWithExtra]
    mutations: Optional[list[Mutation]] = None
    root: "Traceable"
    parent: Optional["Traceable"] = None
    server: TraceServerInterface
    _is_dirty: bool = False

    def _mark_dirty(self) -> None:
        """Recursively mark this object and its ancestors as dirty and removes their refs."""
        self._is_dirty = True
        self.ref = None
        if (
            # Written this way to satisfy mypy
            self.parent is not self
            and self.parent is not None
            and hasattr(self.parent, "_mark_dirty")
        ):
            self.parent._mark_dirty()

    def add_mutation(
        self, path: tuple[str, ...], operation: MutationOperation, *args: Any
    ) -> None:
        if self.mutations is None:
            self.mutations = []
        self.mutations.append(make_mutation(path, operation, args))

    def save(self) -> ObjectRef:
        if not isinstance(self.ref, ObjectRef):
            raise ValueError("Can only save from object refs")
        if self.root is not self:
            raise ValueError("Can only save from root object")
        if self.mutations is None:
            raise ValueError("No mutations to save")

        mutations = self.mutations
        self.mutations = None
        raise NotImplementedError("Traceable.save not implemented")
        # return self.server.mutate(self.ref, mutations)


def pydantic_getattribute(self: BaseModel, name: str) -> Any:
    attribute = object.__getattribute__(self, name)
    if name not in object.__getattribute__(self, "model_fields"):
        return attribute
    if name == "ref":
        try:
            return object.__getattribute__(self, "ref")
        except AttributeError:
            return None

    server = gc.server if (gc := get_weave_client()) else None
    res = attribute_access_result(self, attribute, name, server=server)

    # We need this because we override __getattribute__ and wrap the returned values.
    # The wrapped result may be mutable (e.g. list), so we need to replace the attribute
    # on self so that mutations are applied to the correct object and the user gets back
    # what they expect when they call `self.<name>`.
    self.__dict__[name] = res
    return res


def attribute_access_result(
    self: object,
    val_attr_val: Any,
    attr_name: str,
    *,
    server: Optional[TraceServerInterface],
) -> Any:
    # Not ideal, what about properties?
    if callable(val_attr_val):
        return maybe_bind_method(val_attr_val, self)

    if (ref := getattr(self, "ref", None)) is None:
        # Even if we have not parent ref, if our current value is an object
        # ref, we should still process it with make_trace_obj. Practically,
        # this allows a user to "get" a Model, update a field, then invoke it.
        # Primary test: `test_dirty_model_op_retrieval`
        if not isinstance(val_attr_val, ObjectRef):
            return val_attr_val
        new_ref = None
    else:
        new_ref = ref.with_attr(attr_name)

    if server is None:
        return val_attr_val

    root = getattr(self, "root", None)

    return make_trace_obj(
        val_attr_val,
        new_ref,
        server,
        root,
        self,
    )


class WeaveObject(Traceable):
    def __init__(
        self,
        val: Any,
        ref: Optional[RefWithExtra],
        server: TraceServerInterface,
        root: typing.Optional[Traceable],
        parent: Optional[Traceable] = None,
    ) -> None:
        self._val = val
        self.ref = ref
        self.server = server
        self.root = root or self
        self.parent = parent

    def __getattribute__(self, __name: str) -> Any:
        try:
            return object.__getattribute__(self, __name)
        except AttributeError:
            pass
        val_attr_val = object.__getattribute__(self._val, __name)
        result = attribute_access_result(self, val_attr_val, __name, server=self.server)

        # Store the result on _val so we don't deref next time.
        try:
            object.__setattr__(self._val, __name, result)
        except AttributeError:
            # Happens if self._val.<name> is a property. Return the raw value instead
            # of a Traceable value.
            return val_attr_val
        return result

    def __setattr__(self, __name: str, __value: Any) -> None:
        if __name in [
            "_val",
            "ref",
            "server",
            "root",
            "mutations",
            "_is_dirty",
            "parent",
        ]:
            return object.__setattr__(self, __name, __value)
        else:
            self._mark_dirty()
            if isinstance(__value, Traceable):
                __value.parent = self

            return setattr(self._val, __name, __value)

    def __dir__(self) -> list[str]:
        return dir(self._val)

    def __repr__(self) -> str:
        return f"WeaveObject({self._val})"

    def __eq__(self, other: Any) -> bool:
        return self._val == other


class WeaveTable(Traceable):
    filter: TableRowFilter

    def __init__(
        self,
        table_ref: Optional[TableRef],
        ref: Optional[RefWithExtra],
        server: TraceServerInterface,
        filter: TableRowFilter,
        root: Optional[Traceable],
        parent: Optional[Traceable] = None,
    ) -> None:
        self.table_ref = table_ref
        self.filter = filter
        self.ref = ref  # type: ignore
        self.server = server
        self.root = root or self
        self.parent = parent
        self._rows: Optional[list[dict]] = None

        # _prefetched_rows is a local cache of rows that can be used to
        # avoid a remote call. Should only be used by internal code.
        self._prefetched_rows: Optional[list[dict]] = None

    @property
    def rows(self) -> list[dict]:
        if self._rows is None:
            should_local_iter = (
                self.ref is not None
                and self.table_ref is not None
                and self.table_ref.row_digests is not None
                and self._prefetched_rows is not None
            )
            if should_local_iter:
                self._rows = list(self._local_iter_with_remote_fallback())
            else:
                self._rows = list(self._remote_iter())
        return self._rows

    @rows.setter
    def rows(self, value: list[dict]) -> None:
        if not all(isinstance(row, dict) for row in value):
            raise ValueError("All table rows must be dicts")

        self._rows = value
        self._mark_dirty()

    def set_prefetched_rows(self, prefetched_rows: list[dict]) -> None:
        """Sets the rows to a local cache of rows that can be used to
        avoid a remote call. Should only be used by internal code.

        It is expected that these rows are the exact same rows that would
        be returned by a query for this table. Failing to meet this expectation
        will cause table operations to behave unexpectedly.
        """
        if self._rows is not None:
            raise ValueError(
                "Cannot set prefetched rows on WeaveTable when rows are already loaded"
            )
        self._prefetched_rows = prefetched_rows

    def __len__(self) -> int:
        return len(self.rows)

    def __eq__(self, other: Any) -> bool:
        return self.rows == other

    def _mark_dirty(self) -> None:
        self.table_ref = None
        self._prefetched_rows = None
        super()._mark_dirty()

    def _local_iter_with_remote_fallback(self) -> Generator[dict, None, None]:
        """
        This is the case where we:
        1. Have all the rows in memory
        2. Have all the row digests

        In this case, we don't need to make any calls and can just return the rows
        """
        if (
            self.ref is None
            or self.table_ref is None
            or self.table_ref.row_digests is None
            or self._prefetched_rows is None
        ):
            logger.error(
                "Expected all row digests and prefetched rows to be set, falling back to remote iteration"
            )
            if get_raise_on_captured_errors():
                raise
            yield from self._remote_iter()
            return

        row_digest_len = len(self.table_ref.row_digests)
        prefetched_rows_len = len(self._prefetched_rows)
        if row_digest_len != prefetched_rows_len:
            logger.error(
                f"Expected length of row digests ({row_digest_len}) to match prefetched rows ({prefetched_rows_len}). Falling back to remote iteration."
            )
            if get_raise_on_captured_errors():
                raise
            yield from self._remote_iter()
            return

        for ndx, item in enumerate(self.table_ref.row_digests):
            new_ref = self.ref.with_item(item)
            val = self._prefetched_rows[ndx]
            res = from_json(
                val, self.table_ref.entity + "/" + self.table_ref.project, self.server
            )
            res = make_trace_obj(res, new_ref, self.server, self.root)
            yield res

    def _remote_iter(self) -> Generator[dict, None, None]:
        page_index = 0
        page_size = 1000
        while True:
            if self.table_ref is None:
                break

            response = self.server.table_query(
                TableQueryReq(
                    project_id=f"{self.table_ref.entity}/{self.table_ref.project}",
                    digest=self.table_ref.digest,
                    offset=page_index * page_size,
                    limit=page_size,
                    # filter=self.filter,
                )
            )

            if self._prefetched_rows is not None and len(response.rows) != len(
                self._prefetched_rows
            ):
                logger.error(
                    f"Expected length of response rows ({len(response.rows)}) to match prefetched rows ({len(self._prefetched_rows)}). Ignoring prefetched rows."
                )
                if get_raise_on_captured_errors():
                    raise
                self._prefetched_rows = None

            for ndx, item in enumerate(response.rows):
                new_ref = self.ref.with_item(item.digest) if self.ref else None
                # Here, we use the raw rows if they exist, otherwise we use the
                # rows from the server. This is a temporary trick to ensure
                # we don't re-deserialize the rows on every access. Once all servers
                # return digests, this branch can be removed because anytime we have prefetched
                # rows we should also have the digests - and we should be in the
                #  _local_iter_with_remote_fallback case.
                val = (
                    item.val
                    if self._prefetched_rows is None
                    else self._prefetched_rows[ndx]
                )
                res = from_json(
                    val,
                    self.table_ref.entity + "/" + self.table_ref.project,
                    self.server,
                )
                res = make_trace_obj(res, new_ref, self.server, self.root)
                yield res

            if len(response.rows) < page_size:
                break

            page_index += 1

    def __getitem__(self, key: Union[int, slice, str]) -> Any:
        rows = self.rows
        if isinstance(key, (int, slice)):
            return rows[key]

        for row in rows:
            if row.ref.extra[-1] == key:  # type: ignore
                return row

        raise KeyError(f"Row ID not found: {key}")

    def __iter__(self) -> Iterator[dict]:
        return iter(self.rows)

    def append(self, val: dict) -> None:
        if not isinstance(val, dict):
            raise ValueError("Can only append dicts to tables")
        self._mark_dirty()
        self.rows.append(val)

    def pop(self, index: int) -> None:
        self._mark_dirty()
        self.rows.pop(index)


class WeaveList(Traceable, list):
    def __init__(
        self,
        *args: Any,
        server: TraceServerInterface,
        ref: Optional[RefWithExtra] = None,
        root: Optional[Traceable] = None,
        parent: Optional[Traceable] = None,
    ) -> None:
        self.server = server

        self.ref = ref
        self.root = root if root is not None else self
        self.parent = parent
        super().__init__(*args)

    def __getitem__(self, i: Union[SupportsIndex, slice]) -> Any:
        if isinstance(i, slice):
            raise ValueError("Slices not yet supported")
        index = operator.index(i)
        new_ref = self.ref.with_index(index) if self.ref else None
        index_val = super().__getitem__(index)
        return make_trace_obj(index_val, new_ref, self.server, self.root)

    def __setitem__(self, i: Union[SupportsIndex, slice], value: Any) -> None:
        if isinstance(i, slice):
            raise ValueError("Slices not yet supported")
        if (index := operator.index(i)) >= len(self):
            raise IndexError("list assignment index out of range")

        # Though this ostensibly only marks the parent (list) as dirty, siblings
        # will also get new refs because their old refs are relative to the parent
        # (the element refs will be extras of the new parent ref)
        self._mark_dirty()
        if isinstance(value, Traceable):
            value.parent = self

        super().__setitem__(index, value)

    def append(self, item: Any) -> None:
        self._mark_dirty()
        if isinstance(item, Traceable):
            item.parent = self

        super().append(item)

    def __iter__(self) -> Iterator[Any]:
        for i in range(len(self)):
            yield self[i]

    def __repr__(self) -> str:
        return f"WeaveList({super().__repr__()})"

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, list):
            return False
        if len(self) != len(other):
            return False
        for v1, v2 in zip(self, other):
            if v1 != v2:
                return False
        return True


class WeaveDict(Traceable, dict):
    def __init__(
        self,
        *args: Any,
        server: TraceServerInterface,
        ref: Optional[RefWithExtra] = None,
        root: Optional[Traceable] = None,
        parent: Optional[Traceable] = None,
        **kwargs: Any,
    ) -> None:
        self.server = server

        self.ref = ref
        self.root = root if root is not None else self
        self.parent = parent
        super().__init__(*args, **kwargs)

    def __getitem__(self, key: str) -> Any:
        new_ref = self.ref.with_key(key) if self.ref else None
        v = super().__getitem__(key)
        return make_trace_obj(v, new_ref, self.server, self.root)

    def get(self, key: str, default: Any = None) -> Any:
        new_ref = self.ref.with_key(key) if self.ref else None
        v = super().get(key, default)
        return make_trace_obj(v, new_ref, self.server, self.root)

    def __setitem__(self, key: str, value: Any) -> None:
        # Though this ostensibly only marks the parent (dict) as dirty, siblings
        # will also get new refs because their old refs are relative to the parent
        # (the element refs will be extras of the new parent ref)
        self._mark_dirty()
        if isinstance(value, Traceable):
            value.parent = self

        super().__setitem__(key, value)

    def keys(self) -> Generator[Any, Any, Any]:  # type: ignore
        yield from super().keys()

    def values(self) -> Generator[Any, Any, Any]:  # type: ignore
        for k in self.keys():
            yield self[k]

    def items(self) -> Generator[tuple[Any, Any], Any, Any]:  # type: ignore
        for k in self.keys():
            yield k, self[k]

    def __iter__(self) -> Iterator[str]:
        # Simply define this to so that d = WeaveDict({'a': 1, 'b': 2})); d2 = dict(d)
        # works. The dict(d) constructor works differently if __iter__ is not defined
        # on d.
        return super().__iter__()

    def __repr__(self) -> str:
        return f"WeaveDict({super().__repr__()})"

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, dict):
            return False
        if len(self) != len(other):
            return False
        for k, v in self.items():
            if k not in other:
                return False
            if other[k] != v:
                return False
        return True


def make_trace_obj(
    val: Any,
    new_ref: Optional[RefWithExtra],  # Can this actually be None?
    server: TraceServerInterface,
    root: Optional[Traceable],
    parent: Any = None,
) -> Any:
    if isinstance(val, Traceable):
        # If val is a WeaveTable, we want to refer to it via the outer object
        # that it is within, rather than via the TableRef. For example we
        # want Dataset row refs to be Dataset.rows[id] rather than table[id]
        if isinstance(val, WeaveTable):
            val.ref = new_ref
        return val
    if hasattr(val, "ref") and isinstance(val.ref, RefWithExtra):
        # The Traceable check above does not currently work for Ops, where we
        # directly attach a ref, or to our Boxed classes. We should use Traceable
        # for all of these, but for now we need to check for the ref attribute.
        return val
    # Derefence val and create the appropriate wrapper object
    extra: tuple[str, ...] = ()
    if isinstance(val, ObjectRef):
        new_ref = val
        extra = val.extra
        read_res = server.obj_read(
            ObjReadReq(
                project_id=f"{val.entity}/{val.project}",
                object_id=val.name,
                digest=val.digest,
            )
        )
        val = from_json(read_res.obj.val, val.entity + "/" + val.project, server)

    if isinstance(val, Table):
        val_ref = val.ref
        if not isinstance(val_ref, TableRef):
            val_table_ref = getattr(val, "table_ref", None)
            if not isinstance(val_table_ref, TableRef):
                raise InternalError(
                    "Expected Table.ref or Table.table_ref to be TableRef"
                )
            val_ref = val_table_ref
        rows = val.rows
        val = WeaveTable(
            table_ref=val_ref,
            ref=new_ref,
            server=server,
            filter=TableRowFilter(),
            root=root,
            parent=parent,
        )
        # Use in memory rows! This is the case where we are making
        # a trace object from an existing Table! If we don't do this
        # then the WeaveTable will try to fetch all the rows from the
        # server, throwing away the in memory rows. This is really expensive
        # when we are doing evaluations!
        val.set_prefetched_rows(rows)
    if isinstance(val, TableRef):
        val = WeaveTable(
            table_ref=val,
            ref=new_ref,
            server=server,
            filter=TableRowFilter(),
            root=root,
            parent=parent,
        )

    if extra:
        # This is where extra resolution happens?
        for extra_index in range(0, len(extra), 2):
            op, arg = extra[extra_index], extra[extra_index + 1]
            if op == DICT_KEY_EDGE_NAME:
                val = val[arg]
            elif op == OBJECT_ATTR_EDGE_NAME:
                val = getattr(val, arg)
            elif op == LIST_INDEX_EDGE_NAME:
                val = val[int(arg)]
            elif op == TABLE_ROW_ID_EDGE_NAME:
                val = val[arg]
            else:
                raise ValueError(f"Unknown ref type: {extra[extra_index]}")

            # need to deref if we encounter these
            if isinstance(val, TableRef):
                val = WeaveTable(
                    table_ref=val,
                    ref=new_ref,
                    server=server,
                    filter=TableRowFilter(),
                    root=root,
                    parent=parent,
                )

    if not isinstance(val, Traceable):
        if isinstance(val, ObjectRecord):
            return WeaveObject(
                val, ref=new_ref, server=server, root=root, parent=parent
            )
        elif isinstance(val, list):
            return WeaveList(val, ref=new_ref, server=server, root=root, parent=parent)
        elif isinstance(val, dict):
            return WeaveDict(val, ref=new_ref, server=server, root=root)
    if is_op(val) and inspect.signature(val.resolve_fn).parameters.get("self"):
        # This condition attempts to bind the current `self` to the attribute if
        # it happens to be both an `Op` and have a `self` argument. This is a
        # bit of a hack since we are not always sure that the current object is
        # the correct object to bind. There are 3 cases:
        # 1. The attribute is part of the instance methods and the binding is
        #    correct
        # 2. The attribute is assigned as a property and is not bound at
        #    assignment time. In this case, it is "unlikely" that the args
        #    contain a `self` argument - which is why we apply this heuristic.
        # 3. The attribute is assigned as a property and is bound to another
        #    object at the time of assignment. In this case, the binding is
        #    incorrect. However, in our evaluation use case we do not have this
        #    case. We are accepting the incorrect assignment here for the sake
        #    of expediency, but should be fixed.
        if parent is None:
            raise MissingSelfInstanceError(
                f"{val.name} Op requires a bound self instance. Must be called from an instance method."
            )
        # TODO: This binding is correct but not done for consistency with the
        # not-yet-saved-method-op API which requires explicitly passing self
        # val.call = partial(call, val, parent)
        val = maybe_bind_method(val, parent)
    box_val = box.box(val)
    if isinstance(box_val, pydantic_v1.BaseModel) or is_op(val):
        box_val.__dict__["ref"] = new_ref
    elif box_val is None or isinstance(box_val, bool):
        # We intentionally don't box None and bools because it's imposible to
        # make them behave like the underlying True/False/None objects in python.
        # This is unlike other objects (dict, list, int) that can be inherited
        # from and compared.

        # The tradeoff we're making here is:
        # 1. We won't ref track bools or None when passed into a call; but
        # 2. Users can compare them pythonically (e.g. `x is None` vs. `x == None`)

        pass
    else:
        if hasattr(box_val, "ref"):
            setattr(box_val, "ref", new_ref)
    return box_val


class MissingSelfInstanceError(ValueError):
    pass



# This file contains the definition for "Internal" weave refs. These should never be
# exposed to the user and should only be used internally by the weave-trace service.
# Specifically, the user-space operates on plain-text `entity/project` scopes. While
# internally, we operate on internal `project_id`s scopes. At rest, and in the database,
# we store the internal `project_id`. However, over the wire, we use the plain-text. Practically,
# the trace interface should only ever operate on internal refs.
import dataclasses
import urllib
from typing import Any, Union

WEAVE_INTERNAL_SCHEME = "weave-trace-internal"
WEAVE_SCHEME = "weave"
WEAVE_PRIVATE_SCHEME = "weave-private"

DICT_KEY_EDGE_NAME = "key"
LIST_INDEX_EDGE_NAME = "index"
OBJECT_ATTR_EDGE_NAME = "attr"
TABLE_ROW_ID_EDGE_NAME = "id"

valid_edge_names = (
    DICT_KEY_EDGE_NAME,
    LIST_INDEX_EDGE_NAME,
    OBJECT_ATTR_EDGE_NAME,
    TABLE_ROW_ID_EDGE_NAME,
)


class InvalidInternalRef(ValueError):
    pass


def extra_value_quoter(s: str) -> str:
    # Here, we encode all non alpha-numerics or `_.-~`.
    return urllib.parse.quote(s, safe="")


def validate_extra(extra: list[str]) -> None:
    """The `extra` path is always a series of `edge_name`/`edge_value` pairs.
    For example: `attr/attr_name/index/0`. The `edge_name` is one of the
    following: `key`, `index`, `attr`, `id`. The `edge_value` is a string
    that corresponds to the edge name. In the case of `attr` and `key`, the
    edge value is purely user-defined and can be any string.
    """
    if len(extra) % 2 != 0:
        raise InvalidInternalRef("Extra fields must be key-value pairs.")

    for i, e in enumerate(extra):
        if i % 2 == 0:
            # Here we are in the edge name position
            if e not in (
                DICT_KEY_EDGE_NAME,
                LIST_INDEX_EDGE_NAME,
                OBJECT_ATTR_EDGE_NAME,
                TABLE_ROW_ID_EDGE_NAME,
            ):
                raise InvalidInternalRef(
                    f"Invalid extra edge name at index {i}: {extra}"
                )
        else:
            # Here we are in the edge value position
            # There is only a single rule here:
            if extra[i - 1] == LIST_INDEX_EDGE_NAME:
                try:
                    int(e)
                except ValueError:
                    raise InvalidInternalRef(
                        f"Invalid list edge value at index {i}: {extra}"
                    )
            pass


def validate_no_slashes(s: str, field_name: str) -> None:
    if "/" in s:
        raise InvalidInternalRef(f"{field_name} cannot contain '/'")


def validate_no_colons(s: str, field_name: str) -> None:
    if ":" in s:
        raise InvalidInternalRef(f"{field_name} cannot contain ':'")


@dataclasses.dataclass(frozen=True)
class InternalTableRef:
    project_id: str
    digest: str

    def __post_init__(self) -> None:
        validate_no_slashes(self.project_id, "project_id")
        validate_no_slashes(self.digest, "digest")

    def uri(self) -> str:
        return f"{WEAVE_INTERNAL_SCHEME}:///{self.project_id}/table/{self.digest}"


@dataclasses.dataclass(frozen=True)
class InternalObjectRef:
    project_id: str
    name: str
    version: str
    extra: list[str] = dataclasses.field(default_factory=list)

    def __post_init__(self) -> None:
        validate_no_slashes(self.project_id, "project_id")
        validate_no_slashes(self.version, "version")
        validate_no_colons(self.version, "version")
        validate_extra(self.extra)
        validate_no_slashes(self.name, "name")
        validate_no_colons(self.name, "name")

    def uri(self) -> str:
        u = f"{WEAVE_INTERNAL_SCHEME}:///{self.project_id}/object/{self.name}:{self.version}"
        if self.extra:
            u += "/" + "/".join(extra_value_quoter(e) for e in self.extra)
        return u


@dataclasses.dataclass(frozen=True)
class InternalOpRef(InternalObjectRef):
    def uri(self) -> str:
        u = f"{WEAVE_INTERNAL_SCHEME}:///{self.project_id}/op/{self.name}:{self.version}"
        if self.extra:
            u += "/" + "/".join(extra_value_quoter(e) for e in self.extra)
        return u


@dataclasses.dataclass(frozen=True)
class InternalCallRef:
    project_id: str
    id: str
    extra: list[str] = dataclasses.field(default_factory=list)

    def __post_init__(self) -> None:
        validate_no_slashes(self.project_id, "project_id")
        validate_no_slashes(self.id, "id")
        # Note: we don't actually support extra fields for calls, but when
        # we do, we need to add edge names to the known list
        validate_extra(self.extra)

    def uri(self) -> str:
        u = f"{WEAVE_INTERNAL_SCHEME}:///{self.project_id}/call/{self.id}"
        if self.extra:
            u += "/" + "/".join(extra_value_quoter(e) for e in self.extra)
        return u


def parse_internal_uri(
    uri: str,
) -> Union[InternalObjectRef, InternalTableRef, InternalCallRef]:
    if uri.startswith(f"{WEAVE_INTERNAL_SCHEME}:///"):
        path = uri[len(f"{WEAVE_INTERNAL_SCHEME}:///") :]
        parts = path.split("/")
        if len(parts) < 2:
            raise InvalidInternalRef(f"Invalid URI: {uri}. Must have at least 2 parts")
        project_id, kind = parts[:2]
        remaining = parts[2:]
    elif uri.startswith(f"{WEAVE_SCHEME}:///"):
        path = uri[len(f"{WEAVE_SCHEME}:///") :]
        parts = path.split("/")
        if len(parts) < 3:
            raise InvalidInternalRef(f"Invalid URI: {uri}. Must have at least 3 parts")
        entity, project, kind = parts[:3]
        project_id = f"{entity}/{project}"
        remaining = parts[3:]
    else:
        raise InvalidInternalRef(
            f"Invalid URI: {uri}. Must start with {WEAVE_INTERNAL_SCHEME}:/// or {WEAVE_SCHEME}:///"
        )
    if kind == "table":
        return InternalTableRef(project_id=project_id, digest=remaining[0])
    elif kind == "object":
        name, version, extra = _parse_remaining(remaining)
        return InternalObjectRef(
            project_id=project_id,
            name=name,
            version=version,
            extra=extra,
        )
    elif kind == "op":
        name, version, extra = _parse_remaining(remaining)
        return InternalOpRef(
            project_id=project_id,
            name=name,
            version=version,
            extra=extra,
        )
    elif kind == "call":
        id_ = remaining[0]
        return InternalCallRef(project_id=project_id, id=id_)
    else:
        raise InvalidInternalRef(f"Unknown ref kind: {kind}")


def _parse_remaining(remaining: list[str]) -> tuple[str, str, list[str]]:
    """`remaining` refers to everything after `object` or `op` in the ref.
    It is expected to be pre-split by slashes into parts. The return
    is a tuple of name, version, and extra parts, properly unquoted.
    """
    name, version = remaining[0].split(":")
    extra = remaining[1:]
    if len(extra) == 1 and extra[0] == "":
        extra = []
    else:
        extra = [urllib.parse.unquote(r) for r in extra]

    return name, version, extra


def string_will_be_interpreted_as_ref(s: str) -> bool:
    return s.startswith(f"{WEAVE_INTERNAL_SCHEME}:///") or s.startswith(
        f"{WEAVE_SCHEME}:///"
    )


def any_will_be_interpreted_as_ref_str(val: Any) -> bool:
    return isinstance(val, str) and string_will_be_interpreted_as_ref(val)



import os

import git


def get_git_root(path: str) -> str:
    git_repo = git.Repo(path, search_parent_directories=True)
    git_root = git_repo.git.rev_parse("--show-toplevel")
    return os.path.relpath(git_root, path)


NOTEBOOK_TEST_TEMPLATE = """
import {checkWeaveNotebookOutputs} from '%s/notebooks';

describe('%s notebook test', () => {
    it('passes', () =>
        checkWeaveNotebookOutputs('%s')
    );
});
""".strip()


def makedir(path):
    path = os.path.dirname(path)
    if not os.path.exists(path):
        os.makedirs(path)


def main():
    # For any notebook found within the examples directory, generate a
    # test file in integration_test/cypress/e2e/notebooks.
    git_root = get_git_root(os.curdir)
    examples_dir = os.path.join(git_root, "examples")
    integration_dir = os.path.join(git_root, "integration_test")
    test_dir = os.path.join(integration_dir, "cypress", "e2e", "notebooks")
    for root, dirs, files in os.walk(examples_dir):
        subdir = root.split("/")[-1]
        if subdir == "skip_test" or subdir == "ProductionMonitoring":
            continue
        for file in files:
            if "ipynb_checkpoints" in root:
                continue
            if file.endswith(".ipynb"):
                # keep location of Embeddings.ipynb for ProdMon demos and explore_embeddings for getting started
                # TODO: replace with tested demo notebook
                if "mbedding" in file:
                    continue
                notebook_path = os.path.join(root, file)
                test_path = os.path.join(
                    test_dir,
                    os.path.relpath(root, "examples"),
                    file.replace(".ipynb", ".cy.ts"),
                )
                makedir(test_path)
                cypress_notebook_path = os.path.relpath(notebook_path, integration_dir)
                relative_notebook_path = os.path.relpath(
                    test_dir, os.path.dirname(test_path)
                )
                test_code = NOTEBOOK_TEST_TEMPLATE % (
                    relative_notebook_path,
                    cypress_notebook_path,
                    cypress_notebook_path,
                )
                print("PATHS %s %s\n%s" % (notebook_path, test_path, test_code))
                with open(test_path, "w") as f:
                    f.write(test_code)


if __name__ == "__main__":
    main()



import io
import json
import logging
import typing as t

import tenacity
from pydantic import BaseModel, ValidationError

from weave.legacy.weave.environment import weave_trace_server_url
from weave.legacy.weave.wandb_interface import project_creator
from weave.trace_server import requests
from weave.trace_server import trace_server_interface as tsi
from weave.trace_server.async_batch_processor import AsyncBatchProcessor

logger = logging.getLogger(__name__)


class StartBatchItem(BaseModel):
    mode: str = "start"
    req: tsi.CallStartReq


class EndBatchItem(BaseModel):
    mode: str = "end"
    req: tsi.CallEndReq


class Batch(BaseModel):
    batch: t.List[t.Union[StartBatchItem, EndBatchItem]]


class ServerInfoRes(BaseModel):
    min_required_weave_python_version: str


REMOTE_REQUEST_BYTES_LIMIT = (
    (32 - 1) * 1024 * 1024
)  # 32 MiB (real limit) - 1 MiB (buffer)

REMOTE_REQUEST_RETRY_DURATION = 60 * 60 * 36  # 36 hours
REMOTE_REQUEST_RETRY_MAX_INTERVAL = 60 * 5  # 5 minutes


def _is_retryable_exception(e: Exception) -> bool:
    # Don't retry pydantic validation errors
    if isinstance(e, ValidationError):
        return False

    # Don't retry on HTTP 4xx (except 429)
    if isinstance(e, requests.HTTPError) and e.response is not None:
        code_class = e.response.status_code // 100

        # Bad request, not rate-limiting
        if code_class == 4 and e.response.status_code != 429:
            return False

        # Unknown server error
        # TODO(np): We need to fix the server to return proper status codes
        # for downstream 401, 403, 404, etc... Those should propagate back to
        # the client.
        if e.response.status_code == 500:
            return False

    # Otherwise, retry: Non-500 5xx, OSError, ConnectionError, ConnectionResetError, IOError, etc...
    return True


def _log_retry(retry_state: tenacity.RetryCallState) -> None:
    logger.info(
        "retry_attempt",
        extra={
            "fn": retry_state.fn,
            "attempt_number": retry_state.attempt_number,
            "exception": str(retry_state.outcome.exception()),
        },
    )


def _log_failure(retry_state: tenacity.RetryCallState) -> t.Any:
    logger.info(
        "retry_failed",
        extra={
            "fn": retry_state.fn,
            "attempt_number": retry_state.attempt_number,
            "exception": str(retry_state.outcome.exception()),
        },
    )
    return retry_state.outcome.result()


class RemoteHTTPTraceServer(tsi.TraceServerInterface):
    trace_server_url: str

    # My current batching is not safe in notebooks, disable it for now
    def __init__(
        self,
        trace_server_url: str,
        should_batch: bool = False,
        *,
        remote_request_bytes_limit: int = REMOTE_REQUEST_BYTES_LIMIT,
    ):
        super().__init__()
        self.trace_server_url = trace_server_url
        self.should_batch = should_batch
        if self.should_batch:
            self.call_processor = AsyncBatchProcessor(self._flush_calls)
        self._auth: t.Optional[t.Tuple[str, str]] = None
        self.remote_request_bytes_limit = remote_request_bytes_limit

    def ensure_project_exists(
        self, entity: str, project: str
    ) -> tsi.EnsureProjectExistsRes:
        # TODO: This should happen in the wandb backend, not here, and it's slow
        # (hundreds of ms)
        return tsi.EnsureProjectExistsRes.model_validate(
            project_creator.ensure_project_exists(entity, project)
        )

    @classmethod
    def from_env(cls, should_batch: bool = False) -> "RemoteHTTPTraceServer":
        return cls(weave_trace_server_url(), should_batch)

    def set_auth(self, auth: t.Tuple[str, str]) -> None:
        self._auth = auth

    @tenacity.retry(
        stop=tenacity.stop_after_delay(REMOTE_REQUEST_RETRY_DURATION),
        wait=tenacity.wait_exponential_jitter(
            initial=1, max=REMOTE_REQUEST_RETRY_MAX_INTERVAL
        ),
        retry=tenacity.retry_if_exception(_is_retryable_exception),
        before_sleep=_log_retry,
        retry_error_callback=_log_failure,
        reraise=True,
    )
    def _flush_calls(
        self,
        batch: t.List,
        *,
        _should_update_batch_size: bool = True,
    ) -> None:
        if len(batch) == 0:
            return

        data = Batch(batch=batch).model_dump_json()
        encoded_data = data.encode("utf-8")
        encoded_bytes = len(encoded_data)

        # Update target batch size (this allows us to have a dynamic batch size based on the size of the data being sent)
        estimated_bytes_per_item = encoded_bytes / len(batch)
        if _should_update_batch_size and estimated_bytes_per_item > 0:
            target_batch_size = int(
                self.remote_request_bytes_limit // estimated_bytes_per_item
            )
            self.call_processor.max_batch_size = max(1, target_batch_size)

        # If the batch is too big, recursively split it in half
        if encoded_bytes > self.remote_request_bytes_limit and len(batch) > 1:
            split_idx = int(len(batch) // 2)
            self._flush_calls(batch[:split_idx], _should_update_batch_size=False)
            self._flush_calls(batch[split_idx:], _should_update_batch_size=False)
            return

        r = requests.post(
            self.trace_server_url + "/call/upsert_batch",
            data=encoded_data,
            auth=self._auth,
        )
        if r.status_code == 413:
            # handle 413 explicitly to provide actionable error message
            reason = json.loads(r.text)["reason"]
            raise requests.HTTPError(f"413 Client Error: {reason}", response=r)
        r.raise_for_status()

    @tenacity.retry(
        stop=tenacity.stop_after_delay(REMOTE_REQUEST_RETRY_DURATION),
        wait=tenacity.wait_exponential_jitter(
            initial=1, max=REMOTE_REQUEST_RETRY_MAX_INTERVAL
        ),
        retry=tenacity.retry_if_exception(_is_retryable_exception),
        before_sleep=_log_retry,
        retry_error_callback=_log_failure,
        reraise=True,
    )
    def _generic_request_executor(
        self,
        url: str,
        req: BaseModel,
        stream: bool = False,
    ) -> requests.Response:
        r = requests.post(
            self.trace_server_url + url,
            # `by_alias` is required since we have Mongo-style properties in the
            # query models that are aliased to conform to start with `$`. Without
            # this, the model_dump will use the internal property names which are
            # not valid for the `model_validate` step.
            data=req.model_dump_json(by_alias=True).encode("utf-8"),
            auth=self._auth,
            stream=stream,
        )
        if r.status_code == 500:
            reason_val = r.text
            try:
                reason_val = json.dumps(json.loads(reason_val), indent=2)
            except json.JSONDecodeError:
                reason_val = f"Reason: {reason_val}"
            raise requests.HTTPError(
                f"500 Server Error: Internal Server Error for url: {url}. {reason_val}",
                response=r,
            )
        r.raise_for_status()

        return r

    def _generic_request(
        self,
        url: str,
        req: BaseModel,
        req_model: t.Type[BaseModel],
        res_model: t.Type[BaseModel],
    ) -> BaseModel:
        if isinstance(req, dict):
            req = req_model.model_validate(req)
        r = self._generic_request_executor(url, req)
        return res_model.model_validate(r.json())

    def _generic_stream_request(
        self,
        url: str,
        req: BaseModel,
        req_model: t.Type[BaseModel],
        res_model: t.Type[BaseModel],
    ) -> t.Iterator[BaseModel]:
        if isinstance(req, dict):
            req = req_model.model_validate(req)
        r = self._generic_request_executor(url, req, stream=True)
        for line in r.iter_lines():
            if line:
                yield res_model.model_validate(json.loads(line))

    @tenacity.retry(
        stop=tenacity.stop_after_delay(REMOTE_REQUEST_RETRY_DURATION),
        wait=tenacity.wait_exponential_jitter(
            initial=1, max=REMOTE_REQUEST_RETRY_MAX_INTERVAL
        ),
        retry=tenacity.retry_if_exception(_is_retryable_exception),
        before_sleep=_log_retry,
        retry_error_callback=_log_failure,
        reraise=True,
    )
    def server_info(self) -> ServerInfoRes:
        r = requests.get(self.trace_server_url + "/server_info")
        r.raise_for_status()
        return ServerInfoRes.model_validate(r.json())

    # Call API
    def call_start(
        self, req: t.Union[tsi.CallStartReq, t.Dict[str, t.Any]]
    ) -> tsi.CallStartRes:
        if self.should_batch:
            req_as_obj: tsi.CallStartReq
            if isinstance(req, dict):
                req_as_obj = tsi.CallStartReq.model_validate(req)
            else:
                req_as_obj = req
            if req_as_obj.start.id == None or req_as_obj.start.trace_id == None:
                raise ValueError(
                    "CallStartReq must have id and trace_id when batching."
                )
            self.call_processor.enqueue([StartBatchItem(req=req_as_obj)])
            return tsi.CallStartRes(
                id=req_as_obj.start.id, trace_id=req_as_obj.start.trace_id
            )
        return self._generic_request(
            "/call/start", req, tsi.CallStartReq, tsi.CallStartRes
        )

    def call_end(
        self, req: t.Union[tsi.CallEndReq, t.Dict[str, t.Any]]
    ) -> tsi.CallEndRes:
        if self.should_batch:
            req_as_obj: tsi.CallEndReq
            if isinstance(req, dict):
                req_as_obj = tsi.CallEndReq.model_validate(req)
            else:
                req_as_obj = req
            self.call_processor.enqueue([EndBatchItem(req=req_as_obj)])
            return tsi.CallEndRes()
        return self._generic_request("/call/end", req, tsi.CallEndReq, tsi.CallEndRes)

    def call_read(
        self, req: t.Union[tsi.CallReadReq, t.Dict[str, t.Any]]
    ) -> tsi.CallReadRes:
        return self._generic_request(
            "/call/read", req, tsi.CallReadReq, tsi.CallReadRes
        )

    def calls_query(
        self, req: t.Union[tsi.CallsQueryReq, t.Dict[str, t.Any]]
    ) -> tsi.CallsQueryRes:
        return self._generic_request(
            "/calls/query", req, tsi.CallsQueryReq, tsi.CallsQueryRes
        )

    def calls_query_stream(self, req: tsi.CallsQueryReq) -> t.Iterator[tsi.CallSchema]:
        return self._generic_stream_request(
            "/calls/stream_query", req, tsi.CallsQueryReq, tsi.CallSchema
        )

    def calls_query_stats(
        self, req: t.Union[tsi.CallsQueryStatsReq, t.Dict[str, t.Any]]
    ) -> tsi.CallsQueryStatsRes:
        return self._generic_request(
            "/calls/query_stats", req, tsi.CallsQueryStatsReq, tsi.CallsQueryStatsRes
        )

    def calls_delete(
        self, req: t.Union[tsi.CallsDeleteReq, t.Dict[str, t.Any]]
    ) -> tsi.CallsDeleteRes:
        return self._generic_request(
            "/calls/delete", req, tsi.CallsDeleteReq, tsi.CallsDeleteRes
        )

    def call_update(
        self, req: t.Union[tsi.CallUpdateReq, t.Dict[str, t.Any]]
    ) -> tsi.CallUpdateRes:
        return self._generic_request(
            "/call/update", req, tsi.CallUpdateReq, tsi.CallUpdateRes
        )

    # Op API

    def op_create(
        self, req: t.Union[tsi.OpCreateReq, t.Dict[str, t.Any]]
    ) -> tsi.OpCreateRes:
        return self._generic_request(
            "/op/create", req, tsi.OpCreateReq, tsi.OpCreateRes
        )

    def op_read(self, req: t.Union[tsi.OpReadReq, t.Dict[str, t.Any]]) -> tsi.OpReadRes:
        return self._generic_request("/op/read", req, tsi.OpReadReq, tsi.OpReadRes)

    def ops_query(
        self, req: t.Union[tsi.OpQueryReq, t.Dict[str, t.Any]]
    ) -> tsi.OpQueryRes:
        return self._generic_request("/ops/query", req, tsi.OpQueryReq, tsi.OpQueryRes)

    # Obj API

    def obj_create(
        self, req: t.Union[tsi.ObjCreateReq, t.Dict[str, t.Any]]
    ) -> tsi.ObjCreateRes:
        return self._generic_request(
            "/obj/create", req, tsi.ObjCreateReq, tsi.ObjCreateRes
        )

    def obj_read(
        self, req: t.Union[tsi.ObjReadReq, t.Dict[str, t.Any]]
    ) -> tsi.ObjReadRes:
        return self._generic_request("/obj/read", req, tsi.ObjReadReq, tsi.ObjReadRes)

    def objs_query(
        self, req: t.Union[tsi.ObjQueryReq, t.Dict[str, t.Any]]
    ) -> tsi.ObjQueryRes:
        return self._generic_request(
            "/objs/query", req, tsi.ObjQueryReq, tsi.ObjQueryRes
        )

    def table_create(
        self, req: t.Union[tsi.TableCreateReq, t.Dict[str, t.Any]]
    ) -> tsi.TableCreateRes:
        """Similar to `calls/batch_upsert`, we can dynamically adjust the payload size
        due to the property that table creation can be decomposed into a series of
        updates. This is useful when the table creation size is too big to be sent in
        a single request. We can create an empty table first, then update the table
        with the rows.
        """
        if isinstance(req, dict):
            req = tsi.TableCreateReq.model_validate(req)
        req = t.cast(tsi.TableCreateReq, req)

        estimated_bytes = len(req.model_dump_json(by_alias=True).encode("utf-8"))
        if estimated_bytes > self.remote_request_bytes_limit:
            initialization_req = tsi.TableCreateReq(
                table=tsi.TableSchemaForInsert(
                    project_id=req.table.project_id,
                    rows=[],
                )
            )
            initialization_res = self.table_create(initialization_req)

            update_req = tsi.TableUpdateReq(
                project_id=req.table.project_id,
                base_digest=initialization_res.digest,
                updates=[
                    tsi.TableAppendSpec(append=tsi.TableAppendSpecPayload(row=row))
                    for row in req.table.rows
                ],
            )
            update_res = self.table_update(update_req)

            return tsi.TableCreateRes(
                digest=update_res.digest, row_digests=update_res.updated_row_digests
            )
        else:
            return self._generic_request(
                "/table/create", req, tsi.TableCreateReq, tsi.TableCreateRes
            )

    def table_update(self, req: tsi.TableUpdateReq) -> tsi.TableUpdateRes:
        """Similar to `calls/batch_upsert`, we can dynamically adjust the payload size
        due to the property that table updates can be decomposed into a series of
        updates.
        """
        if isinstance(req, dict):
            req = tsi.TableUpdateReq.model_validate(req)
        req = t.cast(tsi.TableUpdateReq, req)

        estimated_bytes = len(req.model_dump_json(by_alias=True).encode("utf-8"))
        if estimated_bytes > self.remote_request_bytes_limit and len(req.updates) > 1:
            split_ndx = len(req.updates) // 2
            first_half_req = tsi.TableUpdateReq(
                project_id=req.project_id,
                base_digest=req.base_digest,
                updates=req.updates[:split_ndx],
            )
            first_half_res = self.table_update(first_half_req)
            second_half_req = tsi.TableUpdateReq(
                project_id=req.project_id,
                base_digest=first_half_res.digest,
                updates=req.updates[split_ndx:],
            )
            second_half_res = self.table_update(second_half_req)
            all_digests = (
                first_half_res.updated_row_digests + second_half_res.updated_row_digests
            )
            return tsi.TableUpdateRes(
                digest=second_half_res.digest, updated_row_digests=all_digests
            )
        else:
            return self._generic_request(
                "/table/update", req, tsi.TableUpdateReq, tsi.TableUpdateRes
            )

    def table_query(
        self, req: t.Union[tsi.TableQueryReq, t.Dict[str, t.Any]]
    ) -> tsi.TableQueryRes:
        return self._generic_request(
            "/table/query", req, tsi.TableQueryReq, tsi.TableQueryRes
        )

    def refs_read_batch(
        self, req: t.Union[tsi.RefsReadBatchReq, t.Dict[str, t.Any]]
    ) -> tsi.RefsReadBatchRes:
        return self._generic_request(
            "/refs/read_batch", req, tsi.RefsReadBatchReq, tsi.RefsReadBatchRes
        )

    @tenacity.retry(
        stop=tenacity.stop_after_delay(REMOTE_REQUEST_RETRY_DURATION),
        wait=tenacity.wait_exponential_jitter(
            initial=1, max=REMOTE_REQUEST_RETRY_MAX_INTERVAL
        ),
        retry=tenacity.retry_if_exception(_is_retryable_exception),
        before_sleep=_log_retry,
        retry_error_callback=_log_failure,
        reraise=True,
    )
    def file_create(self, req: tsi.FileCreateReq) -> tsi.FileCreateRes:
        r = requests.post(
            self.trace_server_url + "/files/create",
            auth=self._auth,
            data={"project_id": req.project_id},
            files={"file": (req.name, req.content)},
        )
        r.raise_for_status()
        return tsi.FileCreateRes.model_validate(r.json())

    @tenacity.retry(
        stop=tenacity.stop_after_delay(REMOTE_REQUEST_RETRY_DURATION),
        wait=tenacity.wait_exponential_jitter(
            initial=1, max=REMOTE_REQUEST_RETRY_MAX_INTERVAL
        ),
        retry=tenacity.retry_if_exception(_is_retryable_exception),
        before_sleep=_log_retry,
        retry_error_callback=_log_failure,
        reraise=True,
    )
    def file_content_read(self, req: tsi.FileContentReadReq) -> tsi.FileContentReadRes:
        r = requests.post(
            self.trace_server_url + "/files/content",
            json={"project_id": req.project_id, "digest": req.digest},
            auth=self._auth,
        )
        r.raise_for_status()
        # TODO: Should stream to disk rather than to memory
        bytes = io.BytesIO()
        bytes.writelines(r.iter_content())
        bytes.seek(0)
        return tsi.FileContentReadRes(content=bytes.read())

    def feedback_create(
        self, req: t.Union[tsi.FeedbackCreateReq, t.Dict[str, t.Any]]
    ) -> tsi.FeedbackCreateRes:
        return self._generic_request(
            "/feedback/create", req, tsi.FeedbackCreateReq, tsi.FeedbackCreateRes
        )

    def feedback_query(
        self, req: t.Union[tsi.FeedbackQueryReq, t.Dict[str, t.Any]]
    ) -> tsi.FeedbackQueryRes:
        return self._generic_request(
            "/feedback/query", req, tsi.FeedbackQueryReq, tsi.FeedbackQueryRes
        )

    def feedback_purge(
        self, req: t.Union[tsi.FeedbackPurgeReq, t.Dict[str, t.Any]]
    ) -> tsi.FeedbackPurgeRes:
        return self._generic_request(
            "/feedback/purge", req, tsi.FeedbackPurgeReq, tsi.FeedbackPurgeRes
        )

    # Cost API
    def cost_query(
        self, req: t.Union[tsi.CostQueryReq, t.Dict[str, t.Any]]
    ) -> tsi.CostQueryRes:
        return self._generic_request(
            "/cost/query", req, tsi.CostQueryReq, tsi.CostQueryRes
        )

    def cost_create(
        self, req: t.Union[tsi.CostCreateReq, t.Dict[str, t.Any]]
    ) -> tsi.CostCreateRes:
        return self._generic_request(
            "/cost/create", req, tsi.CostCreateReq, tsi.CostCreateRes
        )

    def cost_purge(
        self, req: t.Union[tsi.CostPurgeReq, t.Dict[str, t.Any]]
    ) -> tsi.CostPurgeRes:
        return self._generic_request(
            "/cost/purge", req, tsi.CostPurgeReq, tsi.CostPurgeRes
        )


__docspec__ = [
    RemoteHTTPTraceServer,
    ServerInfoRes,
    StartBatchItem,
    EndBatchItem,
    Batch,
]



"""Settings for Weave.

## `disabled`

* Environment Variable: `WEAVE_DISABLED`
* Settings Key: `disabled`
* Default: `False`
* Type: `bool`

If True, all weave ops will behave like regular functions and no network requests will be made.

## `print_call_link`

* Environment Variable: `WEAVE_PRINT_CALL_LINK`
* Settings Key: `print_call_link`
* Default: `True`
* Type: `bool`

If True, prints a link to the Weave UI when calling a weave op.
"""

import os
from contextvars import ContextVar
from typing import Any, Optional, Union

from pydantic import BaseModel, ConfigDict, PrivateAttr

SETTINGS_PREFIX = "WEAVE_"

# Attention Devs:
# To add new settings:
# 1. Add a new field to `UserSettings`
# 2. Add a new `should_{xyz}` function


class UserSettings(BaseModel):
    """User configuration for Weave.

    All configs can be overrided with environment variables.  The precedence is
    environment variables > `weave.trace.settings.UserSettings`."""

    disabled: bool = False
    """Toggles Weave tracing.
    
    If True, all weave ops will behave like regular functions.
    Can be overrided with the environment variable `WEAVE_DISABLED`"""

    print_call_link: bool = True
    """Toggles link printing to the terminal.

    If True, prints a link to the Weave UI when calling a weave op.
    Can be overrided with the environment variable `WEAVE_PRINT_CALL_LINK`"""

    capture_code: bool = True
    """Toggles code capture for ops.
    
    If True, saves code for ops so they can be reloaded for later use.
    Can be overrided with the environment variable `WEAVE_CAPTURE_CODE`
    
    WARNING: Switching between `save_code=True` and `save_code=False` mid-script
    may lead to unexpected behaviour.  Make sure this is only set once at the start!
    """

    model_config = ConfigDict(extra="forbid")
    _is_first_apply: bool = PrivateAttr(True)

    def _reset(self) -> None:
        for name, field in self.model_fields.items():
            setattr(self, name, field.default)

    def apply(self) -> None:
        if self._is_first_apply:
            self._is_first_apply = False
        else:
            self._reset()

        for name in self.model_fields:
            context_var = _context_vars[name]
            context_var.set(getattr(self, name))


def should_disable_weave() -> bool:
    return _should("disabled")


def should_print_call_link() -> bool:
    return _should("print_call_link")


def should_capture_code() -> bool:
    return _should("capture_code")


def parse_and_apply_settings(
    settings: Optional[Union[UserSettings, dict[str, Any]]] = None,
) -> None:
    if settings is None:
        user_settings = UserSettings()
    if isinstance(settings, dict):
        user_settings = UserSettings.model_validate(settings)
    if isinstance(settings, UserSettings):
        user_settings = settings

    user_settings.apply()


_context_vars = {
    name: ContextVar(name, default=field.default)
    for name, field in UserSettings.model_fields.items()
}


def _str2bool_truthy(v: str) -> bool:
    return v.lower() in ("yes", "true", "1", "on")


def _should(name: str) -> bool:
    if env := os.getenv(f"{SETTINGS_PREFIX}{name.upper()}"):
        return _str2bool_truthy(env)
    return _context_vars[name].get()


__doc_spec__ = [UserSettings]



from weave.trace_server.emoji_util import detone_emojis, detone_shortcode

SHORTCODE_NO_ZWJ_NO_TONE = ":pool_8_ball:"
SHORTCODE_NO_ZWJ_TONE = ":thumbs_up_medium-dark_skin_tone:"
SHORTCODE_ZWJ_ONE_TONE = ":judge_medium-light_skin_tone:"
SHORTCODE_ZWJ_MULTIPLE_TONES = ":handshake_dark_skin_tone_medium_skin_tone:"

EMOJI_8_BALL = "🎱"
EMOJI_THUMBS_UP_TONE = "👍🏾"
EMOJI_THUMBS_UP_NO_TONE = "👍"
EMOJI_JUDGE_TONE = "🧑🏼‍⚖️"
EMOJI_JUDGE_NO_TONE = "🧑‍⚖️"
EMOJI_HANDSHAKE_TONES = "🫱🏿‍🫲🏽"
EMOJI_HANDSHAKE_NO_TONES = "🤝"

SENTENCE_TONES = f"The {EMOJI_JUDGE_TONE} had us {EMOJI_HANDSHAKE_TONES}."
SENTENCE_NO_TONES = f"The {EMOJI_JUDGE_NO_TONE} had us {EMOJI_HANDSHAKE_NO_TONES}."


def test_detone_shortcode() -> None:
    assert detone_shortcode(SHORTCODE_NO_ZWJ_NO_TONE) == ":pool_8_ball:"
    assert detone_shortcode(SHORTCODE_NO_ZWJ_TONE) == ":thumbs_up:"
    assert detone_shortcode(SHORTCODE_ZWJ_ONE_TONE) == ":judge:"
    assert detone_shortcode(SHORTCODE_ZWJ_MULTIPLE_TONES) == ":handshake:"


def test_detone_emoji() -> None:
    assert detone_emojis(EMOJI_8_BALL) == EMOJI_8_BALL
    assert detone_emojis(EMOJI_THUMBS_UP_TONE) == EMOJI_THUMBS_UP_NO_TONE
    assert detone_emojis(EMOJI_JUDGE_TONE) == EMOJI_JUDGE_NO_TONE
    assert detone_emojis(EMOJI_HANDSHAKE_TONES) == EMOJI_HANDSHAKE_NO_TONES
    assert detone_emojis(SENTENCE_TONES) == SENTENCE_NO_TONES



import pyarrow as pa
import pyarrow.compute as pc

from weave.legacy.weave import weave_types as types
from weave.legacy.weave.arrow.list_ import ArrowWeaveList, ArrowWeaveListType
from weave.legacy.weave.decorator_arrow_op import arrow_op
from weave.legacy.weave.decorator_op import op
from weave.legacy.weave.ops_arrow import util

nullable_binary_input_type = {
    "self": ArrowWeaveListType(types.optional(types.RefType())),
    "other": types.UnionType(
        types.optional(types.RefType()),
        ArrowWeaveListType(types.optional(types.RefType())),
    ),
}


@arrow_op(
    name="ArrowWeaveListRef-equal",
    input_type=nullable_binary_input_type,
    output_type=ArrowWeaveListType(types.Boolean()),
)
def ref_equal(self, other):
    from weave.legacy.weave import storage

    # Weave engine automatically derefs, so we need to undo that via
    # _get_ref here.
    other_ref = storage._get_ref(other)
    if not other_ref:
        # if there's not a ref, maybe this is a vector of refs?
        if isinstance(other, ArrowWeaveList):
            other = other._arrow_data
        else:
            # Otherwise... this shouldn't happen I think?
            raise ValueError(
                f"Cannot compare ArrowWeaveListRef to {type(other)}: {other}"
            )
    other = str(other_ref)
    self_ = self.map_column(
        lambda col, path: ArrowWeaveList(col._arrow_data, types.String())
    )
    # self_arrow_data = pc.cast(self._arrow_data, pa.string())
    result = util.equal(self_._arrow_data, other)
    return ArrowWeaveList(result, types.Boolean(), self._artifact)


@arrow_op(
    name="ArrowWeaveListRef-toUri",
    input_type={
        "self": ArrowWeaveListType(types.optional(types.RefType())),
    },
    output_type=ArrowWeaveListType(types.optional(types.String())),
)
def ref_to_uri(self):
    return self.map_column(
        lambda col, path: ArrowWeaveList(col._arrow_data, types.String())
    )



import pytest

from weave.legacy.weave.ops_primitives.string import _levenshtein


@pytest.mark.parametrize(
    "str1,str2,expected",
    [
        ["", "", 0],
        ["a", "a", 0],
        ["", "a", 1],
        ["a", "b", 1],
        ["a", "ab", 1],
        ["ab", "ac", 1],
        ["ab", "ba", 2],
        ["", "ab", 2],
        ["ab", "cd", 2],
        ["abc", "cab", 2],
        ["", "abc", 3],
        ["abc", "def", 3],
    ],
)
def test_levenshtein(str1, str2, expected):
    assert _levenshtein(str1, str2) == expected
    assert _levenshtein(str2, str1) == expected



import asyncio
import contextlib
import time

import numpy as np
import pytest
import requests

from weave.legacy.weave import api as weave
from weave.legacy.weave import client as _client
from weave.legacy.weave import context_state, ops
from weave.legacy.weave import server as _server
from weave.legacy.weave import weave_types as types
from weave.legacy.weave.decorator_op import op
from weave.legacy.weave.weave_internal import make_const_node

SERVER_TYPES = ["inprocess", "subprocess", "http"]


@contextlib.contextmanager
def client(server_type):
    if server_type == "inprocess":
        yield _client.Client(_server.InProcessServer())
    elif server_type == "subprocess":
        subprocess_server = _server.SubprocessServerClient()
        yield _client.Client(subprocess_server)
        subprocess_server.shutdown()
    elif server_type == "http":
        port = 9995
        server = _server.HttpServer(port)
        server.start()
        yield _client.Client(_server.HttpServerClient("http://127.0.0.1:%s" % port))
        server.shutdown()


@pytest.mark.parametrize("server_type", SERVER_TYPES)
@pytest.mark.timeout(10)
def test_basic(server_type):
    with client(server_type) as wc:
        nine = make_const_node(weave.types.Number(), 9)
        assert weave.use(nine + 3, client=wc) == 12


@pytest.mark.parametrize("server_type", SERVER_TYPES)
@pytest.mark.timeout(10)
def test_type_returning_op(server_type, cereal_csv):
    with client(server_type) as wc:
        csv_type = weave.use(ops.local_path_refine_type(cereal_csv), client=wc)
        assert csv_type.name == "local_file"


@pytest.mark.timeout(10)
def test_500_does_raise_jsondecode_error_from_http_server():
    with client("http") as wc:

        @op()
        def custom_op_that_should_return_500(x: str) -> str:
            if x == "abcd":
                raise ValueError("returning 500")
            return x + "a"

        # should not raise json decoder error, but an HTTP error insteard
        with pytest.raises(requests.exceptions.HTTPError):
            weave.use(custom_op_that_should_return_500("abcd"), client=wc)
